# Generated by Django 2.2.4 on 2019-11-01 22:46

import django.core.validators
from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone
import model_utils.fields
import uuid


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Asset',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('code', models.TextField(default='USD', validators=[django.core.validators.MinLengthValidator(1)])),
                ('issuer', models.TextField(default='GCXEIFQV5CAFTMDWW3YYUBMRXHQEE3DWQCGQLJMC4PUK6GGWERICUTOQ', validators=[django.core.validators.MinLengthValidator(56)])),
                ('deposit_enabled', models.BooleanField(default=True)),
                ('deposit_fee_fixed', models.FloatField(blank=True, default=1.0)),
                ('deposit_fee_percent', models.FloatField(blank=True, default=0.01)),
                ('deposit_min_amount', models.FloatField(blank=True, default=10.0)),
                ('deposit_max_amount', models.FloatField(blank=True, default=10000.0)),
                ('withdrawal_enabled', models.BooleanField(default=True)),
                ('withdrawal_fee_fixed', models.FloatField(blank=True, default=1.0)),
                ('withdrawal_fee_percent', models.FloatField(blank=True, default=0.01)),
                ('withdrawal_min_amount', models.FloatField(blank=True, default=10.0)),
                ('withdrawal_max_amount', models.FloatField(blank=True, default=10000.0)),
            ],
        ),
        migrations.CreateModel(
            name='Transaction',
            fields=[
                ('id', models.UUIDField(default=uuid.uuid4, primary_key=True, serialize=False)),
                ('stellar_account', models.TextField(validators=[django.core.validators.MinLengthValidator(1)])),
                ('kind', models.CharField(choices=[('deposit', 'deposit'), ('withdrawal', 'withdrawal')], default='deposit', max_length=20)),
                ('status', models.CharField(choices=[('completed', 'completed'), ('pending_external', 'pending_external'), ('pending_anchor', 'pending_anchor'), ('pending_stellar', 'pending_stellar'), ('pending_trust', 'pending_trust'), ('pending_user', 'pending_user'), ('pending_user_transfer_start', 'pending_user_transfer_start'), ('incomplete', 'incomplete'), ('no_market', 'no_market'), ('too_small', 'too_small'), ('too_large', 'too_large'), ('error', 'error')], default='pending_external', max_length=30)),
                ('status_eta', models.IntegerField(blank=True, default=3600, null=True)),
                ('stellar_transaction_id', models.TextField(blank=True, null=True)),
                ('external_transaction_id', models.TextField(blank=True, null=True)),
                ('amount_in', models.FloatField(blank=True, null=True)),
                ('amount_out', models.FloatField(blank=True, null=True)),
                ('amount_fee', models.FloatField(blank=True, null=True)),
                ('started_at', models.DateTimeField(auto_now_add=True)),
                ('completed_at', models.DateTimeField(null=True)),
                ('from_address', models.TextField(blank=True, null=True)),
                ('to_address', models.TextField(blank=True, null=True)),
                ('external_extra', models.TextField(blank=True, null=True)),
                ('external_extra_text', models.TextField(blank=True, null=True)),
                ('deposit_memo', models.TextField(blank=True, null=True)),
                ('deposit_memo_type', models.CharField(choices=[('text', 'text'), ('id', 'id'), ('hash', 'hash')], default='text', max_length=10)),
                ('withdraw_anchor_account', models.TextField(blank=True, null=True)),
                ('withdraw_memo', models.TextField(blank=True, null=True)),
                ('withdraw_memo_type', models.CharField(choices=[('text', 'text'), ('id', 'id'), ('hash', 'hash')], default='text', max_length=10)),
                ('asset', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='polaris.Asset')),
            ],
            options={
                'ordering': ('-started_at',),
            },
        ),
    ]



"""This module defines the logic for the `/transaction` endpoint."""
import json
from urllib.parse import urlencode

from rest_framework.decorators import api_view, renderer_classes
from rest_framework.response import Response
from rest_framework.renderers import TemplateHTMLRenderer
from rest_framework import status

from polaris import settings
from django.urls import reverse
from django.views.decorators.clickjacking import xframe_options_exempt

from polaris.helpers import render_error_response, validate_sep10_token, validate_jwt_request
from polaris.models import Transaction
from polaris.transaction.serializers import TransactionSerializer


def _validate_limit(limit):
    limit = int(limit or settings.DEFAULT_PAGE_SIZE)
    if limit < 1:
        raise ValueError
    return limit


def _compute_qset_filters(req_params, translation_dict):
    """
    _compute_qset_filters translates the keys of req_params to the keys of translation_dict.
    If the key isn't present in filters_dict, it is discarded.
    """

    return {
        translation_dict[rp]: req_params[rp]
        for rp in filter(lambda i: i in translation_dict, req_params.keys())
    }


def _get_transaction_from_request(request):
    translation_dict = {
        "id": "id",
        "stellar_transaction_id": "stellar_transaction_id",
        "external_transaction_id": "external_transaction_id",
    }

    qset_filter = _compute_qset_filters(request.GET, translation_dict)
    if not qset_filter:
        raise AttributeError(
            "at least one of id, stellar_transaction_id, or external_transaction_id must be provided"
        )
    try:
        return Transaction.objects.get(**qset_filter)
    except Transaction.DoesNotExist as exc:
        raise exc


def _construct_more_info_url(request):
    qparams_dict = {}
    transaction_id = request.GET.get("id")
    if transaction_id:
        qparams_dict["id"] = transaction_id
    stellar_transaction_id = request.GET.get("stellar_transaction_id")
    if stellar_transaction_id:
        qparams_dict["stellar_transaction_id"] = stellar_transaction_id
    external_transaction_id = request.GET.get("external_transaction_id")
    if external_transaction_id:
        qparams_dict["external_transaction_id"] = external_transaction_id

    qparams = urlencode(qparams_dict)
    path = reverse("more_info")
    path_params = f"{path}?{qparams}"
    return request.build_absolute_uri(path_params)


@xframe_options_exempt
@api_view()
@renderer_classes([TemplateHTMLRenderer])
def more_info(request):
    """
    Popup to display more information about a specific transaction.
    See table: https://github.com/stellar/stellar-protocol/blob/master/ecosystem/sep-0024.md#4-customer-information-status
    """
    try:
        request_transaction = _get_transaction_from_request(request)
    except AttributeError as exc:
        return render_error_response(str(exc), content_type="text/html")
    except Transaction.DoesNotExist:
        return render_error_response(
            "transaction not found",
            status_code=status.HTTP_404_NOT_FOUND,
            content_type="text/html"
        )

    serializer = TransactionSerializer(
        request_transaction,
        context={"more_info_url": _construct_more_info_url(request)},
    )
    tx_json = json.dumps({"transaction": serializer.data})
    return Response(
        {
            "tx_json": tx_json,
            "transaction": request_transaction,
            "asset_code": request_transaction.asset.code,
        },
        template_name="transaction/more_info.html"
    )


@api_view()
@validate_sep10_token()
def transactions(request):
    """
    Definition of the /transactions endpoint, in accordance with SEP-0024.
    See: https://github.com/stellar/stellar-protocol/blob/master/ecosystem/sep-0024.md#transaction-history
    """
    try:
        limit = _validate_limit(request.GET.get("limit"))
    except ValueError:
        return render_error_response(
            "invalid limit",
            status_code=status.HTTP_400_BAD_REQUEST
        )

    if not request.GET.get("asset_code") or not request.GET.get("account"):
        return render_error_response(
            "asset_code and account are required fields",
            status_code=status.HTTP_400_BAD_REQUEST,
        )

    translation_dict = {
        "asset_code": "asset__code",
        "account": "stellar_account",
        "no_older_than": "started_at__gte",
        "kind": "kind",
    }

    qset_filter = _compute_qset_filters(request.GET, translation_dict)

    # Since the Transaction IDs are UUIDs, rather than in the chronological
    # order of their creation, we map the paging ID (if provided) to the
    # started_at field of a Transaction.
    paging_id = request.GET.get("paging_id")
    if paging_id:
        try:
            start_transaction = Transaction.objects.get(id=paging_id)
        except Transaction.DoesNotExist:
            return render_error_response(
                "invalid paging_id", status_code=status.HTTP_400_BAD_REQUEST
            )
        qset_filter["started_at__lt"] = start_transaction.started_at

    transactions_qset = Transaction.objects.filter(**qset_filter)[:limit]
    serializer = TransactionSerializer(
        transactions_qset,
        many=True,
        context={"more_info_url": _construct_more_info_url(request)},
    )

    return Response({"transactions": serializer.data})


@api_view()
@validate_sep10_token()
def transaction(request):
    """
    Definition of the /transaction endpoint, in accordance with SEP-0024.
    See: https://github.com/stellar/stellar-protocol/blob/master/ecosystem/sep-0024.md#single-historical-transaction
    """
    try:
        request_transaction = _get_transaction_from_request(request)
    except AttributeError as exc:
        return render_error_response(str(exc), status_code=status.HTTP_400_BAD_REQUEST)
    except Transaction.DoesNotExist:
        return render_error_response(
            "transaction not found", status_code=status.HTTP_404_NOT_FOUND
        )
    serializer = TransactionSerializer(
        request_transaction,
        context={"more_info_url": _construct_more_info_url(request)},
    )
    return Response({"transaction": serializer.data})



"""This module defines the URL patterns for the `/transaction(s)` endpoints."""
from django.urls import path
from polaris.transaction.views import more_info, transaction, transactions

urlpatterns = [
    path("transaction", transaction),
    path("transactions", transactions),
    path("transaction/more_info", more_info, name="more_info"),
]



"""This module tests the `/transaction/more_info` endpoint."""
import json
import pytest


@pytest.mark.django_db
def test_more_info_required_fields(client, acc1_usd_deposit_transaction_factory):
    """Fails if no required fields are provided."""
    acc1_usd_deposit_transaction_factory()
    response = client.get(f"/transaction/more_info", follow=True)
    assert response.status_code == 400


@pytest.mark.django_db
def test_more_info_id_filter(client, acc1_usd_deposit_transaction_factory):
    """Succeeds if a valid transaction ID is provided."""
    deposit = acc1_usd_deposit_transaction_factory()
    response = client.get(f"/transaction/more_info?id={deposit.id}", follow=True)

    assert response.status_code == 200


@pytest.mark.django_db
def test_more_info_stellar_filter(client, acc1_usd_deposit_transaction_factory):
    """Succeeds if a valid Stellar transaction ID is provided."""
    deposit = acc1_usd_deposit_transaction_factory()
    deposit.stellar_transaction_id = "test_stellar_id"
    deposit.save()
    response = client.get(
        f"/transaction/more_info?stellar_transaction_id={deposit.stellar_transaction_id}",
        follow=True,
    )
    assert response.status_code == 200


@pytest.mark.django_db
def test_more_info_external_filter(client, acc1_usd_deposit_transaction_factory):
    """Succeeds if a valid external transaction ID is provided."""
    deposit = acc1_usd_deposit_transaction_factory()
    response = client.get(
        f"/transaction/more_info?external_transaction_id={deposit.external_transaction_id}",
        follow=True,
    )
    assert response.status_code == 200


@pytest.mark.django_db
def test_more_info_multiple_filters(client, acc1_usd_deposit_transaction_factory):
    """Succeeds if a valid combination of IDs is provided."""
    deposit = acc1_usd_deposit_transaction_factory()
    deposit.stellar_transaction_id = "test_stellar_id"
    deposit.save()
    response = client.get(
        f"/transaction/more_info?id={deposit.id}&external_transaction_id={deposit.external_transaction_id}&stellar_transaction_id={deposit.stellar_transaction_id}",
        follow=True,
    )
    assert response.status_code == 200


@pytest.mark.django_db
def test_more_info_no_result(
    client,
    acc1_usd_deposit_transaction_factory,
    acc2_eth_withdrawal_transaction_factory,
):
    """Fails if an invalid combination of IDs is provided."""
    deposit = acc1_usd_deposit_transaction_factory()
    withdrawal = acc2_eth_withdrawal_transaction_factory()
    response = client.get(
        f"/transaction/more_info?id={deposit.id}&external_transaction_id={withdrawal.external_transaction_id}&stellar_transaction_id={withdrawal.stellar_transaction_id}",
        follow=True,
    )
    assert response.status_code == 404




"""This module defines the models for the polaris app."""
import uuid

from django.contrib import admin
from polaris import settings
from django.core.validators import MinLengthValidator
from django.db import models
from model_utils.models import TimeStampedModel
from model_utils import Choices


class Asset(TimeStampedModel):
    """
    This defines an Asset, as described in the SEP-24 `info` endpoint.
    See: https://github.com/stellar/stellar-protocol/blob/master/ecosystem/sep-0024.md#info
    """

    code = models.TextField(validators=[MinLengthValidator(1)], default="USD")
    issuer = models.TextField(
        validators=[MinLengthValidator(56)], default=settings.STELLAR_ISSUER_ACCOUNT_ADDRESS
    )

    # Deposit-related info
    deposit_enabled = models.BooleanField(null=False, default=True)
    deposit_fee_fixed = models.FloatField(default=1.0, blank=True)
    deposit_fee_percent = models.FloatField(default=0.01, blank=True)
    deposit_min_amount = models.FloatField(default=10.0, blank=True)
    deposit_max_amount = models.FloatField(default=10000.0, blank=True)

    # Withdrawal-related info
    withdrawal_enabled = models.BooleanField(null=False, default=True)
    withdrawal_fee_fixed = models.FloatField(default=1.0, blank=True)
    withdrawal_fee_percent = models.FloatField(default=0.01, blank=True)
    withdrawal_min_amount = models.FloatField(default=10.0, blank=True)
    withdrawal_max_amount = models.FloatField(default=10000.0, blank=True)

    class Meta:
        app_label = "polaris"


class Transaction(models.Model):
    """
    This defines a Transaction, as described in the SEP-24 `transaction` endpoint.
    See: https://github.com/stellar/stellar-protocol/blob/master/ecosystem/sep-0024.md#transactions
    """

    KIND = Choices("deposit", "withdrawal")
    STATUS = Choices(
        "completed",
        "pending_external",
        "pending_anchor",
        "pending_stellar",
        "pending_trust",
        "pending_user",
        "pending_user_transfer_start",
        "incomplete",
        "no_market",
        "too_small",
        "too_large",
        "error",
    )
    MEMO_TYPES = Choices("text", "id", "hash")

    id = models.UUIDField(primary_key=True, default=uuid.uuid4)
    # Stellar account to watch, and asset that is being transactioned
    # NOTE: these fields should not be publicly exposed
    stellar_account = models.TextField(validators=[MinLengthValidator(1)])
    asset = models.ForeignKey("Asset", on_delete=models.CASCADE)

    # These fields can be shown through an API:
    kind = models.CharField(choices=KIND, default=KIND.deposit, max_length=20)
    status = models.CharField(
        choices=STATUS, default=STATUS.pending_external, max_length=30
    )
    status_eta = models.IntegerField(null=True, blank=True, default=3600)
    stellar_transaction_id = models.TextField(null=True, blank=True)
    external_transaction_id = models.TextField(null=True, blank=True)
    amount_in = models.FloatField(null=True, blank=True)
    amount_out = models.FloatField(null=True, blank=True)
    amount_fee = models.FloatField(null=True, blank=True)
    started_at = models.DateTimeField(auto_now_add=True)
    completed_at = models.DateTimeField(null=True)

    from_address = models.TextField(
        null=True, blank=True
    )  # Using from_address since `from` is a reserved keyword
    to_address = models.TextField(
        null=True, blank=True
    )  # Using to_address for naming consistency
    external_extra = models.TextField(null=True, blank=True)
    external_extra_text = models.TextField(null=True, blank=True)
    deposit_memo = models.TextField(null=True, blank=True)
    deposit_memo_type = models.CharField(
        choices=MEMO_TYPES, default=MEMO_TYPES.text, max_length=10
    )
    withdraw_anchor_account = models.TextField(null=True, blank=True)
    withdraw_memo = models.TextField(null=True, blank=True)
    withdraw_memo_type = models.CharField(
        choices=MEMO_TYPES, default=MEMO_TYPES.text, max_length=10
    )

    def asset_name(self):
        return self.asset.code + ":" + self.asset.issuer

    class Meta:
        ordering = ("-started_at",)
        app_label = "polaris"


"""
This module implements the logic for the authentication endpoint, as per SEP 10.
This defines a standard way for wallets and anchors to create authenticated web sessions
on behalf of a user who holds a Stellar account.

See: https://github.com/stellar/stellar-protocol/blob/master/ecosystem/sep-0010.md
"""
import binascii
import json
import time
import jwt

from polaris import settings
from django.http import JsonResponse
from rest_framework import status
from rest_framework.decorators import api_view

from stellar_sdk.transaction_envelope import TransactionEnvelope
from stellar_sdk.sep.stellar_web_authentication import build_challenge_transaction, verify_challenge_transaction
from stellar_sdk.sep.exceptions import InvalidSep10ChallengeError

MIME_URLENCODE, MIME_JSON = "application/x-www-form-urlencoded", "application/json"
ANCHOR_NAME = "SEP 24 Reference"


def _challenge_transaction(client_account):
    """
    Generate the challenge transaction for a client account.
    This is used in `GET <auth>`, as per SEP 10.
    Returns the XDR encoding of that transaction.
    """
    challenge_tx_xdr = build_challenge_transaction(server_secret=settings.STELLAR_DISTRIBUTION_ACCOUNT_SEED,
                                                   client_account_id=client_account,
                                                   anchor_name=ANCHOR_NAME,
                                                   network_passphrase=settings.STELLAR_NETWORK_PASSPHRASE)
    return challenge_tx_xdr


def _get_transaction_urlencode(body):
    """Get the transaction for URL encoded transaction data to `POST <auth>`."""
    body = body.decode("utf-8")
    # Confirm there is only one URL parameter.
    body_arr = body.split("&")
    if len(body_arr) != 1:
        return 0, "multiple query params provided"
    # The format of the transaction parameter key-value pair should be
    # `transaction=<AAA...>`.
    transaction_param = body_arr[0]
    [key, value] = transaction_param.split("<")
    key = key[:-1]  # Remove trailing `=`.
    if key != "transaction":
        return 0, "no transaction provided"
    envelope_xdr = value[:-1]  # Remove trailing `>`.
    return 1, envelope_xdr


def _get_transaction_json(body):
    """Get the transaction for JSON-encoded transaction data to `POST <auth>`."""
    try:
        body_dict = json.loads(body)
    except TypeError:
        return 0, "invalid json"
    try:
        envelope_xdr = body_dict["transaction"]
    except KeyError:
        return 0, "no transaction found"
    return 1, envelope_xdr


def _transaction_from_post_request(request):
    """Get the transaction (base64 XDR) from the `POST <auth>` request."""
    content_type = request.content_type

    if content_type not in [MIME_URLENCODE, MIME_JSON]:
        return 0, "invalid content type"

    # Handle URL-encoded data.
    if content_type == MIME_URLENCODE:
        return _get_transaction_urlencode(request.body)

    # Handle JSON-encoded data.
    return _get_transaction_json(request.body)


def _validate_envelope_xdr(envelope_xdr):
    """
    Validate the provided TransactionEnvelope XDR (base64 string). Return the
    appropriate error if it fails, else the empty string.
    """
    try:
        verify_challenge_transaction(challenge_transaction=envelope_xdr,
                                     server_account_id=settings.STELLAR_DISTRIBUTION_ACCOUNT_ADDRESS,
                                     network_passphrase=settings.STELLAR_NETWORK_PASSPHRASE)
    except InvalidSep10ChallengeError as err:
        return str(err)
    return ""


def _generate_jwt(request, envelope_xdr):
    """
    Generates the JSON web token from the challenge transaction XDR.

    See: https://github.com/stellar/stellar-protocol/blob/master/ecosystem/sep-0010.md#token
    """
    issued_at = time.time()
    hash_hex = binascii.hexlify(
        TransactionEnvelope.from_xdr(envelope_xdr, network_passphrase=settings.STELLAR_NETWORK_PASSPHRASE).hash()
    ).decode()
    jwt_dict = {
        "iss": request.build_absolute_uri("/auth"),
        "sub": settings.STELLAR_DISTRIBUTION_ACCOUNT_ADDRESS,
        "iat": issued_at,
        "exp": issued_at + 24 * 60 * 60,
        "jti": hash_hex,
    }
    encoded_jwt = jwt.encode(jwt_dict, settings.SERVER_JWT_KEY, algorithm="HS256")
    return encoded_jwt.decode("ascii")


def _get_auth(request):
    account = request.GET.get("account")
    if not account:
        return JsonResponse(
            {"error": "no 'account' provided"}, status=status.HTTP_400_BAD_REQUEST
        )
    transaction = _challenge_transaction(account)
    return JsonResponse(
        {"transaction": transaction, "network_passphrase": settings.STELLAR_NETWORK_PASSPHRASE}
    )


def _post_auth(request):
    success, xdr_or_error = _transaction_from_post_request(request)
    if not success:
        return JsonResponse({"error": xdr_or_error}, status=status.HTTP_400_BAD_REQUEST)
    envelope_xdr = xdr_or_error
    validate_error = _validate_envelope_xdr(envelope_xdr)
    if validate_error != "":
        return JsonResponse(
            {"error": validate_error}, status=status.HTTP_400_BAD_REQUEST
        )
    return JsonResponse({"token": _generate_jwt(request, envelope_xdr)})


@api_view(["GET", "POST"])
def auth(request):
    """
    `GET /auth` can be used to get an invalid challenge Stellar transaction. The client
    can then sign it using their private key and hit `POST /auth` to receive a JSON web
    token. That token can be used to authenticate calls to the other SEP 24 endpoints.
    """
    if request.method == "POST":
        return _post_auth(request)
    return _get_auth(request)



"""Helper functions to use across tests."""
import json

from django.conf import settings
from django.http import JsonResponse
from stellar_sdk import Account
from stellar_sdk.client.response import Response
from stellar_sdk.exceptions import NotFoundError


def mock_check_auth_success(request, func):
    """Mocks `helpers.check_auth`, for success."""
    return func(request)


def mock_load_not_exist_account(account_id):
    if account_id != settings.STELLAR_ISSUER_ACCOUNT_ADDRESS and account_id != settings.STELLAR_DISTRIBUTION_ACCOUNT_ADDRESS:
        raise NotFoundError(response=Response(status_code=404, headers={}, url="", text=json.dumps(dict(status=404))))
    return Account(account_id, 1)



from django.contrib import admin
from polaris.models import Transaction


class TransactionAdmin(admin.ModelAdmin):
    """
    This defines the admin view of a Transaction.
    """

    list_display = "id", "asset_name", "kind", "status", "started_at"


admin.site.register(Transaction, TransactionAdmin)



"""This module defines helpers for various endpoints."""
import codecs
import time
import uuid

from polaris import settings
import jwt
from rest_framework import status
from rest_framework.response import Response


from polaris.models import Asset, Transaction


def calc_fee(asset: Asset, operation: str, amount: float) -> float:
    """Calculates fees for an operation with a given asset and amount."""
    if operation == settings.OPERATION_WITHDRAWAL:
        fee_percent = asset.withdrawal_fee_percent
        fee_fixed = asset.withdrawal_fee_fixed
    else:
        fee_percent = asset.deposit_fee_percent
        fee_fixed = asset.deposit_fee_fixed

    # Note (Alex C, 2019-07-12):
    # `op_type` is not used in this context, since there is no fee variation
    # based on operation type in this example implementation, but that can
    # occur in real-life applications.
    return fee_fixed + (fee_percent / 100.0) * amount


def render_error_response(description: str,
                          status_code: int = status.HTTP_400_BAD_REQUEST,
                          content_type: str = "application/json") -> Response:
    """
    Renders an error response in Django.

    Currently supports HTML or JSON responses.
    """
    resp_data = {
        "data": {"error": description, "status_code": status_code},
        "status": status_code,
        "content_type": content_type
    }
    if content_type == "text/html":
        resp_data["template_name"] = "error.html"
    return Response(**resp_data)


def create_transaction_id():
    """Creates a unique UUID for a Transaction, via checking existing entries."""
    while True:
        transaction_id = uuid.uuid4()
        if not Transaction.objects.filter(id=transaction_id).exists():
            break
    return transaction_id


def format_memo_horizon(memo):
    """
    Formats a hex memo, as in the Transaction model, to match
    the base64 Horizon response.
    """
    return (codecs.encode(codecs.decode(memo, "hex"), "base64").decode("utf-8")).strip()


def check_auth(request, func):
    """
    Check SEP 10 authentication in a request.
    Else call the original view function.
    """
    jwt_error_str = validate_jwt_request(request)
    if jwt_error_str:
        return render_error_response(jwt_error_str)
    return func(request)


def validate_sep10_token():
    """Decorator to validate the SEP 10 token in a request."""

    def decorator(view):
        def wrapper(request, *args, **kwargs):
            return check_auth(request, view)

        return wrapper

    return decorator


def validate_jwt_request(request):
    """
    Validate the JSON web token in a request.
    Return the appropriate error string, or empty string if valid.
    """
    # While the SEP 24 spec calls the authorization header "Authorization", django middleware
    # renames this as "HTTP_AUTHORIZATION". We check this header for the JWT.
    jwt_header = request.META.get("HTTP_AUTHORIZATION")
    if not jwt_header:
        return "JWT must be passed as 'Authorization' header"
    if "Bearer" not in jwt_header:
        return "'Authorization' header must be formatted as 'Bearer <token>'"
    encoded_jwt = jwt_header.split(" ")[1]
    if not encoded_jwt:
        return "'jwt' is required"

    # Validate the JWT contents.
    jwt_dict = jwt.decode(encoded_jwt, settings.SERVER_JWT_KEY, algorithms=["HS256"])
    if jwt_dict["iss"] != request.build_absolute_uri("/auth"):
        return "'jwt' has incorrect 'issuer'"
    if jwt_dict["sub"] != settings.STELLAR_DISTRIBUTION_ACCOUNT_ADDRESS:
        return "'jwt' has incorrect 'subject'"
    current_time = time.time()
    if current_time < jwt_dict["iat"] or current_time > jwt_dict["exp"]:
        return "'jwt' is no longer valid"
    # TODO: Investigate if we can validate the JTI, a hex-encoded transaction hash.
    return ""



"""
This module implements the logic for the `/withdraw` endpoint. This lets a user
withdraw some asset from their Stellar account into a non Stellar based asset.
"""
import json
import uuid
from urllib.parse import urlencode

from polaris import settings
from django.shortcuts import render
from django.urls import reverse
from django.views.decorators.clickjacking import xframe_options_exempt
from rest_framework import status
from rest_framework.decorators import api_view, renderer_classes
from rest_framework.response import Response
from rest_framework.renderers import TemplateHTMLRenderer, JSONRenderer

from polaris.helpers import (
    render_error_response,
    create_transaction_id,
    calc_fee,
    validate_sep10_token,
)
from polaris.models import Asset, Transaction
from polaris.transaction.serializers import TransactionSerializer
from polaris.withdraw.forms import WithdrawForm


def _construct_interactive_url(request, asset_code, transaction_id):
    """Constructs the URL for the interactive application for withdrawal info.
    This is located at `/transactions/withdraw/webapp`."""
    qparams = urlencode(
        {"asset_code": asset_code, "transaction_id": transaction_id}
    )
    path = reverse("interactive_withdraw")
    url_params = f"{path}?{qparams}"
    return request.build_absolute_uri(url_params)


def _construct_more_info_url(request):
    """Constructs the more info URL for a withdraw."""
    qparams_dict = {"id": request.GET.get("transaction_id")}
    qparams = urlencode(qparams_dict)
    path = reverse("more_info")
    path_params = f"{path}?{qparams}"
    return request.build_absolute_uri(path_params)


@xframe_options_exempt
@api_view(["GET", "POST"])
@renderer_classes([TemplateHTMLRenderer])
def interactive_withdraw(request):
    """
    `GET /transactions/withdraw/webapp` opens a form used to input information about
    the withdrawal. This creates a corresponding transaction in our database.
    """
    transaction_id = request.GET.get("transaction_id")
    if not transaction_id:
        return render_error_response("no 'transaction_id' provided", content_type="text/html")

    asset_code = request.GET.get("asset_code")
    if not asset_code or not Asset.objects.filter(code=asset_code).exists():
        return render_error_response("invalid 'asset_code'", content_type="text/html")

    # GET: The server needs to display the form for the user to input withdrawal information.
    if request.method == "GET":
        form = WithdrawForm()

    # POST: The user submitted a form with the withdrawal info.
    else:
        if Transaction.objects.filter(id=transaction_id).exists():
            return render_error_response(
                "transaction with matching 'transaction_id' already exists",
                content_type="text/html"
            )
        form = WithdrawForm(request.POST)
        asset = Asset.objects.get(code=asset_code)
        form.asset = asset

        # If the form is valid, we create a transaction pending user action
        # and render the success page.
        if form.is_valid():
            amount_in = form.cleaned_data["amount"]
            amount_fee = calc_fee(asset, settings.OPERATION_WITHDRAWAL, amount_in)

            # We use the transaction ID as a memo on the Stellar transaction for the
            # payment in the withdrawal. This lets us identify that as uniquely
            # corresponding to this `Transaction` in the database. But a UUID4 is a 32
            # character hex string, while the Stellar HashMemo requires a 64 character
            # hex-encoded (32 byte) string. So, we zero-pad the ID to create an
            # appropriately sized string for the `HashMemo`.
            transaction_id_hex = uuid.UUID(transaction_id).hex
            withdraw_memo = "0" * (64 - len(transaction_id_hex)) + transaction_id_hex
            transaction = Transaction(
                id=transaction_id,
                stellar_account=settings.STELLAR_DISTRIBUTION_ACCOUNT_ADDRESS,
                asset=asset,
                kind=Transaction.KIND.withdrawal,
                status=Transaction.STATUS.pending_user_transfer_start,
                amount_in=amount_in,
                amount_fee=amount_fee,
                withdraw_anchor_account=settings.STELLAR_DISTRIBUTION_ACCOUNT_ADDRESS,
                withdraw_memo=withdraw_memo,
                withdraw_memo_type=Transaction.MEMO_TYPES.hash,
            )
            transaction.save()

            serializer = TransactionSerializer(
                transaction,
                context={"more_info_url": _construct_more_info_url(request)},
            )
            tx_json = json.dumps({"transaction": serializer.data})
            return Response(
                {
                    "tx_json": tx_json,
                    "transaction": transaction,
                    "asset_code": asset_code,
                },
                template_name="transaction/more_info.html"
            )
    return Response({"form": form}, template_name="withdraw/form.html")


@api_view(["POST"])
@validate_sep10_token()
@renderer_classes([JSONRenderer])
def withdraw(request):
    """
    `POST /withdraw` initiates the withdrawal and returns an interactive
    withdrawal form to the user.
    """
    asset_code = request.POST.get("asset_code")
    if not asset_code:
        return render_error_response("'asset_code' is required")

    # TODO: Verify optional arguments.

    # Verify that the asset code exists in our database, with withdraw enabled.
    asset = Asset.objects.filter(code=asset_code).first()
    if not asset or not asset.withdrawal_enabled:
        return render_error_response(f"invalid operation for asset {asset_code}")

    transaction_id = create_transaction_id()
    url = _construct_interactive_url(request, asset_code, transaction_id)
    return Response(
        {"type": "interactive_customer_info_needed", "url": url, "id": transaction_id},
    )



"""This module defines the logic for the `/info` endpoint."""
import json
from rest_framework.decorators import api_view
from rest_framework.response import Response

from polaris import settings
from polaris.models import Asset


def _get_asset_deposit_info(asset: Asset):
    if asset.deposit_enabled:
        return {
            "enabled": True,
            "fee_fixed": asset.deposit_fee_fixed,
            "fee_percent": asset.deposit_fee_percent,
            "min_amount": asset.deposit_min_amount,
            "max_amount": asset.deposit_max_amount,
        }

    return {"enabled": False}


def _get_asset_withdrawal_info(asset: Asset):
    if asset.withdrawal_enabled:
        return {
            "enabled": True,
            "fee_fixed": asset.withdrawal_fee_fixed,
            "fee_percent": asset.withdrawal_fee_percent,
            "min_amount": asset.withdrawal_min_amount,
            "max_amount": asset.withdrawal_max_amount,
        }

    return {"enabled": False}


@api_view()
def info(request):
    """
    Definition of the /info endpoint, in accordance with SEP-0024.
    See: https://github.com/stellar/stellar-protocol/blob/master/ecosystem/sep-0024.md#info
    """

    info_data = {
        "deposit": {},
        "withdraw": {},
        "fee": {"enabled": True},
        "transactions": {"enabled": True},
        "transaction": {"enabled": True},
    }

    for asset in (
        Asset.objects.all()
        .prefetch_related("deposit_fields", "withdrawal_types")
        .iterator()
    ):
        info_data["deposit"][asset.code] = _get_asset_deposit_info(asset)
        info_data["withdraw"][asset.code] = _get_asset_withdrawal_info(asset)

    return Response(info_data)



import logging

from django.utils.timezone import now
from django.core.management import BaseCommand

from stellar_sdk.exceptions import BaseHorizonError
from stellar_sdk.transaction_builder import TransactionBuilder

from polaris import settings
from polaris.models import Transaction


TRUSTLINE_FAILURE_XDR = "AAAAAAAAAGT/////AAAAAQAAAAAAAAAB////+gAAAAA="
SUCCESS_XDR = "AAAAAAAAAGQAAAAAAAAAAQAAAAAAAAABAAAAAAAAAAA="

logger = logging.getLogger(__name__)


class Command(BaseCommand):
    """
    Create and submit the Stellar transaction for the deposit.
    """
    def add_arguments(self, parser):
        parser.add_argument("transaction_id")

    def handle(self, *args, **options):
        self.create_stellar_deposit(options["transaction_id"])
    
    def create_stellar_deposit(self, transaction_id):
        """Create and submit the Stellar transaction for the deposit."""
        transaction = Transaction.objects.get(id=transaction_id)

        # We check the Transaction status to avoid double submission of a Stellar
        # transaction. The Transaction can be either `pending_anchor` if the task
        # is called from `GET transactions/deposit/confirm_transaction` or `pending_trust` if called
        # from the `check_trustlines()`.
        if transaction.status not in [
            Transaction.STATUS.pending_anchor,
            Transaction.STATUS.pending_trust,
        ]:
            logger.debug(
                "unexpected transaction status %s at top of create_stellar_deposit",
                transaction.status,
            )
            return
        transaction.status = Transaction.STATUS.pending_stellar
        transaction.save()

        # We can assume transaction has valid stellar_account, amount_in, and asset
        # because this task is only called after those parameters are validated.
        stellar_account = transaction.stellar_account
        payment_amount = round(transaction.amount_in - transaction.amount_fee, 7)
        asset = transaction.asset.code

        # If the given Stellar account does not exist, create
        # the account with at least enough XLM for the minimum
        # reserve and a trust line (recommended 2.01 XLM), update
        # the transaction in our internal database, and return.

        server = settings.HORIZON_SERVER
        starting_balance = settings.ACCOUNT_STARTING_BALANCE
        server_account = server.load_account(settings.STELLAR_DISTRIBUTION_ACCOUNT_ADDRESS)
        base_fee = server.fetch_base_fee()
        builder = TransactionBuilder(source_account=server_account,
                                    network_passphrase=settings.STELLAR_NETWORK_PASSPHRASE,
                                    base_fee=base_fee)
        try:
            server.load_account(stellar_account)
        except BaseHorizonError as address_exc:
            # 404 code corresponds to Resource Missing.
            if address_exc.status != 404:
                logger.debug(
                    "error with message %s when loading stellar account",
                    address_exc.message,
                )
                return
            transaction_envelope = builder \
                .append_create_account_op(destination=stellar_account,
                                        starting_balance=starting_balance,
                                        source=settings.STELLAR_DISTRIBUTION_ACCOUNT_ADDRESS) \
                .build()
            transaction_envelope.sign(settings.STELLAR_DISTRIBUTION_ACCOUNT_SEED)
            try:
                server.submit_transaction(transaction_envelope)
            except BaseHorizonError as submit_exc:
                logger.debug(
                    f"error with message {submit_exc.message} when submitting create account to horizon"
                )
                return
            transaction.status = Transaction.STATUS.pending_trust
            transaction.save()
            return

        # If the account does exist, deposit the desired amount of the given
        # asset via a Stellar payment. If that payment succeeds, we update the
        # transaction to completed at the current time. If it fails due to a
        # trustline error, we update the database accordingly. Else, we do not update.

        transaction_envelope = builder \
            .append_payment_op(destination=stellar_account,
                            asset_code=asset,
                            asset_issuer=settings.STELLAR_ISSUER_ACCOUNT_ADDRESS,
                            amount=str(payment_amount)) \
            .build()
        transaction_envelope.sign(settings.STELLAR_DISTRIBUTION_ACCOUNT_SEED)
        try:
            response = server.submit_transaction(transaction_envelope)
        # Functional errors at this stage are Horizon errors.
        except BaseHorizonError as exception:
            if TRUSTLINE_FAILURE_XDR not in exception.result_xdr:
                logger.debug(
                    "error with message %s when submitting payment to horizon, non-trustline failure",
                    exception.message,
                )
                return
            logger.debug("trustline error when submitting transaction to horizon")
            transaction.status = Transaction.STATUS.pending_trust
            transaction.save()
            return

        # If this condition is met, the Stellar payment succeeded, so we
        # can mark the transaction as completed.
        if response["result_xdr"] != SUCCESS_XDR:
            logger.debug("payment stellar transaction failed when submitted to horizon")
            return

        transaction.stellar_transaction_id = response["hash"]
        transaction.status = Transaction.STATUS.completed
        transaction.completed_at = now()
        transaction.status_eta = 0  # No more status change.
        transaction.amount_out = payment_amount
        transaction.save()



"""This module defines forms used in the deposit view."""
from django import forms


class DepositForm(forms.Form):
    """This form accepts the amount to deposit from the user."""

    amount = forms.FloatField(
        help_text="Enter the amount to deposit, as a two decimal places float.",
        min_value=0,
        widget=forms.NumberInput(attrs={"class": "input"}),
    )
    asset = None

    def clean_amount(self):
        """Validate the provided amount of an asset."""
        amount = round(self.cleaned_data["amount"], 2)
        if self.asset:
            if amount < self.asset.deposit_min_amount:
                raise forms.ValidationError(
                    f"Amount is below minimum for asset {self.asset.code}"
                )
            elif amount > self.asset.deposit_max_amount:
                raise forms.ValidationError(
                    f"Amount is above maximum for asset {self.asset.code}"
                )
        return amount



"""
This module implements the logic for the `/deposit` endpoint. This lets a user
initiate a deposit of some asset into their Stellar account.

Note that before the Stellar transaction is submitted, an external agent must
confirm that the first step of the deposit successfully completed.
"""
import base64
import binascii
import json
from urllib.parse import urlencode

from polaris import settings
from django.http import JsonResponse
from django.shortcuts import render
from django.urls import reverse
from django.views.decorators.clickjacking import xframe_options_exempt
from django.core.management import call_command
from rest_framework import status
from rest_framework.decorators import api_view, renderer_classes
from rest_framework.response import Response
from rest_framework.renderers import TemplateHTMLRenderer, JSONRenderer
from stellar_sdk.keypair import Keypair
from stellar_sdk.exceptions import Ed25519PublicKeyInvalidError

from polaris.helpers import (
    calc_fee,
    render_error_response,
    create_transaction_id,
    validate_jwt_request,
    validate_sep10_token,
)
from polaris.models import Asset, Transaction
from polaris.transaction.serializers import TransactionSerializer

from polaris.deposit.forms import DepositForm


def _construct_interactive_url(request, asset_code, account, transaction_id): 
    """Constructs the URL for the deposit application for deposit info.
    This is located at `/transactions/deposit/webapp`."""
    qparams = urlencode(
        {
            "asset_code": asset_code,
            "account": account,
            "transaction_id": transaction_id,
        }
    )
    path = reverse("interactive_deposit")
    url_params = f"{path}?{qparams}"
    return request.build_absolute_uri(url_params)


def _construct_more_info_url(request):
    """Constructs the more info URL for a deposit."""
    qparams_dict = {"id": request.GET.get("transaction_id")}
    qparams = urlencode(qparams_dict)
    path = reverse("more_info")
    path_params = f"{path}?{qparams}"
    return request.build_absolute_uri(path_params)


# TODO: The interactive pop-up will be used to retrieve additional info,
# so we should not need to validate these parameters. Alternately, we can
# pass these to the pop-up.
def _verify_optional_args(request):
    """Verify the optional arguments to `GET /deposit`."""
    memo_type = request.POST.get("memo_type")
    if memo_type and memo_type not in ("text", "id", "hash"):
        return render_error_response("invalid 'memo_type'")

    memo = request.POST.get("memo")
    if memo_type and not memo:
        return render_error_response("'memo_type' provided with no 'memo'")

    if memo and not memo_type:
        return render_error_response("'memo' provided with no 'memo_type'")

    if memo_type == "hash":
        try:
            base64.b64encode(base64.b64decode(memo))
        except binascii.Error:
            return render_error_response("'memo' does not match memo_type' hash")
    return None


@api_view()
def confirm_transaction(request):
    """
    `GET /transactions/deposit/confirm_transaction` is used by an external agent to confirm
    that they have processed the transaction. This triggers submission of the
    corresponding Stellar transaction.

    Note that this endpoint is not part of the SEP 24 workflow, it is merely
    a mechanism for confirming the external transaction for demonstration purposes.
    If reusing this technique in a real-life scenario, add a strictly secure
    authentication system.
    """
    # Validate the provided transaction_id and amount.
    transaction_id = request.GET.get("transaction_id")
    if not transaction_id:
        return render_error_response("no 'transaction_id' provided")

    transaction = Transaction.objects.filter(id=transaction_id).first()
    if not transaction:
        return render_error_response(
            "no transaction with matching 'transaction_id' exists"
        )

    amount_str = request.GET.get("amount")
    if not amount_str:
        return render_error_response("no 'amount' provided")
    try:
        amount = float(amount_str)
    except ValueError:
        return render_error_response("non-float 'amount' provided")

    if transaction.amount_in != amount:
        return render_error_response(
            "incorrect 'amount' value for transaction with given 'transaction_id'"
        )

    external_transaction_id = request.GET.get("external_transaction_id")

    # The external deposit has been completed, so the transaction
    # status must now be updated to pending_anchor.
    transaction.status = Transaction.STATUS.pending_anchor
    transaction.status_eta = 5  # Ledger close time.
    transaction.external_transaction_id = external_transaction_id
    transaction.save()
    serializer = TransactionSerializer(
        transaction, context={"more_info_url": _construct_more_info_url(request)}
    )

    # launch the deposit Stellar transaction.
    call_command("create_stellar_deposit", transaction.id)
    return Response({"transaction": serializer.data})


@xframe_options_exempt
@api_view(["GET", "POST"])
@renderer_classes([TemplateHTMLRenderer])
def interactive_deposit(request):
    """
    `GET /transactions/deposit/webapp` opens a form used to input information
    about the deposit. This creates a corresponding transaction in our
    database, pending processing by the external agent.
    """
    # Validate query parameters: account, asset_code, transaction_id.
    account = request.GET.get("account")
    if not account:
        return render_error_response("no 'account' provided", content_type="text/html")

    asset_code = request.GET.get("asset_code")
    if not asset_code or not Asset.objects.filter(code=asset_code).exists():
        return render_error_response("invalid 'asset_code'", content_type="text/html")

    transaction_id = request.GET.get("transaction_id")
    if not transaction_id:
        return render_error_response("no 'transaction_id' provided", content_type="text/html")

    # GET: The server needs to display the form for the user to input the deposit information.
    if request.method == "GET":
        form = DepositForm()
    # POST: The user submitted a form with the amount to deposit.
    else:
        if Transaction.objects.filter(id=transaction_id).exists():
            return render_error_response(
                "transaction with matching 'transaction_id' already exists",
                content_type="text/html"
            )
        form = DepositForm(request.POST)
        asset = Asset.objects.get(code=asset_code)
        form.asset = asset
        # If the form is valid, we create a transaction pending external action
        # and render the success page.
        if form.is_valid():
            amount_in = form.cleaned_data["amount"]
            amount_fee = calc_fee(asset, settings.OPERATION_DEPOSIT, amount_in)
            transaction = Transaction(
                id=transaction_id,
                stellar_account=account,
                asset=asset,
                kind=Transaction.KIND.deposit,
                status=Transaction.STATUS.pending_user_transfer_start,
                amount_in=amount_in,
                amount_fee=amount_fee,
                to_address=account,
            )
            transaction.save()

            serializer = TransactionSerializer(
                transaction,
                context={"more_info_url": _construct_more_info_url(request)},
            )
            tx_json = json.dumps({"transaction": serializer.data})
            return Response(
                {
                    "tx_json": tx_json,
                    "transaction": transaction,
                    "asset_code": transaction.asset.code,
                },
                template_name="transaction/more_info.html",
            )
    return Response({"form": form}, template_name="deposit/form.html")


@api_view(["POST"])
@renderer_classes([JSONRenderer])
@validate_sep10_token()
def deposit(request):
    """
    `POST /transactions/deposit/interactive` initiates the deposit and returns an interactive
    deposit form to the user.
    """
    asset_code = request.POST.get("asset_code")
    stellar_account = request.POST.get("account")

    # Verify that the request is valid.
    if not all([asset_code, stellar_account]):
        return render_error_response(
            "`asset_code` and `account` are required parameters"
        )

    # Verify that the asset code exists in our database, with deposit enabled.
    asset = Asset.objects.filter(code=asset_code).first()
    if not asset or not asset.deposit_enabled:
        return render_error_response(f"invalid operation for asset {asset_code}")

    try:
        Keypair.from_public_key(stellar_account)
    except Ed25519PublicKeyInvalidError:
        return render_error_response("invalid 'account'")

    # Verify the optional request arguments.
    verify_optional_args = _verify_optional_args(request)
    if verify_optional_args:
        return verify_optional_args

    # Construct interactive deposit pop-up URL.
    transaction_id = create_transaction_id()
    url = _construct_interactive_url(request, asset_code, stellar_account, transaction_id)
    return Response(
        {"type": "interactive_customer_info_needed", "url": url, "id": transaction_id},
        status=status.HTTP_403_FORBIDDEN,
    )



"""This module tests the `/withdraw` endpoint."""
import json
from unittest.mock import patch

import pytest
from stellar_sdk.keypair import Keypair
from stellar_sdk.transaction_envelope import TransactionEnvelope

from polaris import settings
from polaris.helpers import format_memo_horizon
from polaris.management.commands.watch_transactions import process_withdrawal
from polaris.models import Transaction
from polaris.tests.helpers import mock_check_auth_success

WITHDRAW_PATH = "/transactions/withdraw/interactive"

@pytest.mark.django_db
@patch("polaris.helpers.check_auth", side_effect=mock_check_auth_success)
def test_withdraw_success(mock_check, client, acc1_usd_withdrawal_transaction_factory):
    """`GET /withdraw` succeeds with no optional arguments."""
    del mock_check
    acc1_usd_withdrawal_transaction_factory()
    response = client.post(WITHDRAW_PATH, {"asset_code": "USD"}, follow=True)
    content = json.loads(response.content)
    assert content["type"] == "interactive_customer_info_needed"


@pytest.mark.django_db
@patch("polaris.helpers.check_auth", side_effect=mock_check_auth_success)
def test_withdraw_invalid_asset(
    mock_check, client, acc1_usd_withdrawal_transaction_factory
):
    """`GET /withdraw` fails with an invalid asset argument."""
    del mock_check
    acc1_usd_withdrawal_transaction_factory()
    response = client.post(WITHDRAW_PATH, {"asset_code": "ETH"}, follow=True)
    content = json.loads(response.content)
    assert response.status_code == 400
    assert content == {"error": "invalid operation for asset ETH", "status_code": 400}


@pytest.mark.django_db
@patch("polaris.helpers.check_auth", side_effect=mock_check_auth_success)
def test_withdraw_no_asset(mock_check, client):
    """`GET /withdraw fails with no asset argument."""
    del mock_check
    response = client.post(WITHDRAW_PATH, follow=True)
    content = json.loads(response.content)
    assert response.status_code == 400
    assert content == {"error": "'asset_code' is required", "status_code": 400}


@pytest.mark.django_db
@patch("polaris.helpers.check_auth", side_effect=mock_check_auth_success)
def test_withdraw_interactive_no_txid(
    mock_check, client, acc1_usd_withdrawal_transaction_factory
):
    """
    `GET /transactions/withdraw/webapp` fails with no transaction_id.
    """
    del mock_check
    acc1_usd_withdrawal_transaction_factory()
    response = client.get(f"/transactions/withdraw/webapp?", follow=True)
    assert response.status_code == 400


@pytest.mark.django_db
@patch("polaris.helpers.check_auth", side_effect=mock_check_auth_success)
def test_withdraw_interactive_no_asset(
    mock_check, client, acc1_usd_withdrawal_transaction_factory
):
    """
    `GET /transactions/withdraw/webapp` fails with no asset_code.
    """
    del mock_check
    acc1_usd_withdrawal_transaction_factory()
    response = client.get(
        f"/transactions/withdraw/webapp?transaction_id=2", follow=True
    )
    assert response.status_code == 400


@pytest.mark.django_db
@patch("polaris.helpers.check_auth", side_effect=mock_check_auth_success)
def test_withdraw_interactive_invalid_asset(
    mock_check, client, acc1_usd_withdrawal_transaction_factory
):
    """
    `GET /transactions/withdraw/webapp` fails with invalid asset_code.
    """
    del mock_check
    acc1_usd_withdrawal_transaction_factory()
    response = client.get(
        f"/transactions/withdraw/webapp?transaction_id=2&asset_code=ETH", follow=True
    )
    assert response.status_code == 400


# TODO: Decompose the below tests, since they call the same logic. The issue: Pytest complains
# about decomposition when passing fixtures to a helper function.


@pytest.mark.django_db
@patch(
    "polaris.management.commands.watch_transactions.stream_transactions",
    return_value=[{}],
)
@patch("polaris.helpers.check_auth", side_effect=mock_check_auth_success)
def test_withdraw_interactive_failure_no_memotype(
    mock_check, mock_transactions, client, acc1_usd_withdrawal_transaction_factory
):
    """
    `GET /transactions/withdraw/webapp` fails with no `memo_type` in Horizon response.
    """
    del mock_check, mock_transactions
    acc1_usd_withdrawal_transaction_factory()
    response = client.post(WITHDRAW_PATH, {"asset_code": "USD"}, follow=True)
    content = json.loads(response.content)
    assert content["type"] == "interactive_customer_info_needed"

    transaction_id = content["id"]
    url = content["url"]
    response = client.post(
        url, {"amount": 20, "bank_account": "123456", "bank": "Bank"}
    )
    assert response.status_code == 200
    assert (
        Transaction.objects.get(id=transaction_id).status
        == Transaction.STATUS.pending_user_transfer_start
    )


@pytest.mark.django_db
@patch(
    "polaris.management.commands.watch_transactions.stream_transactions",
    return_value=[{"memo_type": "not_hash"}],
)
@patch("polaris.helpers.check_auth", side_effect=mock_check_auth_success)
def test_withdraw_interactive_failure_incorrect_memotype(
    mock_check, mock_transactions, client, acc1_usd_withdrawal_transaction_factory
):
    """
    `GET /transactions/withdraw/webapp` fails with incorrect `memo_type` in Horizon response.
    """
    del mock_check, mock_transactions
    acc1_usd_withdrawal_transaction_factory()
    response = client.post(WITHDRAW_PATH, {"asset_code": "USD"}, follow=True)
    content = json.loads(response.content)
    assert content["type"] == "interactive_customer_info_needed"

    transaction_id = content["id"]
    url = content["url"]
    response = client.post(
        url, {"amount": 20, "bank_account": "123456", "bank": "Bank"}
    )
    assert response.status_code == 200
    assert (
        Transaction.objects.get(id=transaction_id).status
        == Transaction.STATUS.pending_user_transfer_start
    )


@pytest.mark.django_db
@patch(
    "polaris.management.commands.watch_transactions.stream_transactions",
    return_value=[{"memo_type": "hash"}],
)
@patch("polaris.helpers.check_auth", side_effect=mock_check_auth_success)
def test_withdraw_interactive_failure_no_memo(
    mock_check, mock_transactions, client, acc1_usd_withdrawal_transaction_factory
):
    """
    `GET /transactions/withdraw/webapp` fails with no `memo` in Horizon response.
    """
    del mock_check, mock_transactions
    acc1_usd_withdrawal_transaction_factory()
    response = client.post(WITHDRAW_PATH, {"asset_code": "USD"}, follow=True)
    content = json.loads(response.content)
    assert content["type"] == "interactive_customer_info_needed"

    transaction_id = content["id"]
    url = content["url"]
    response = client.post(
        url, {"amount": 20, "bank_account": "123456", "bank": "Bank"}
    )
    assert response.status_code == 200
    assert (
        Transaction.objects.get(id=transaction_id).status
        == Transaction.STATUS.pending_user_transfer_start
    )


@pytest.mark.django_db
@patch(
    "polaris.management.commands.watch_transactions.stream_transactions",
    return_value=[{"memo_type": "hash", "memo": "wrong_memo"}],
)
@patch("polaris.helpers.check_auth", side_effect=mock_check_auth_success)
def test_withdraw_interactive_failure_incorrect_memo(
    mock_check, mock_transactions, client, acc1_usd_withdrawal_transaction_factory
):
    """
    `GET /transactions/withdraw/webapp` fails with incorrect `memo` in Horizon response.
    """
    del mock_check, mock_transactions
    acc1_usd_withdrawal_transaction_factory()
    response = client.post(WITHDRAW_PATH, {"asset_code": "USD"}, follow=True)
    content = json.loads(response.content)
    assert content["type"] == "interactive_customer_info_needed"

    transaction_id = content["id"]
    url = content["url"]
    response = client.post(
        url, {"amount": 20, "bank_account": "123456", "bank": "Bank"}
    )
    assert response.status_code == 200
    assert (
        Transaction.objects.get(id=transaction_id).status
        == Transaction.STATUS.pending_user_transfer_start
    )


@pytest.mark.django_db
@patch("polaris.helpers.check_auth", side_effect=mock_check_auth_success)
def test_withdraw_interactive_success_transaction_unsuccessful(
    mock_check, client, acc1_usd_withdrawal_transaction_factory
):
    """
    `GET /transactions/withdraw/webapp` changes transaction to `pending_stellar`
    with unsuccessful transaction.
    """
    del mock_check
    acc1_usd_withdrawal_transaction_factory()
    response = client.post(WITHDRAW_PATH, {"asset_code": "USD"}, follow=True)
    content = json.loads(response.content)
    assert content["type"] == "interactive_customer_info_needed"

    transaction_id = content["id"]
    url = content["url"]
    response = client.post(
        url, {"amount": 50, "bank_account": "123456", "bank": "Bank"}
    )
    assert response.status_code == 200
    transaction = Transaction.objects.get(id=transaction_id)
    assert transaction.status == Transaction.STATUS.pending_user_transfer_start

    withdraw_memo = transaction.withdraw_memo
    mock_response = {
        "memo_type": "hash",
        "memo": format_memo_horizon(withdraw_memo),
        "successful": False,
        "id": "c5e8ada72c0e3c248ac7e1ec0ec97e204c06c295113eedbe632020cd6dc29ff8",
        "envelope_xdr": "AAAAAEU1B1qeJrucdqkbk1mJsnuFaNORfrOAzJyaAy1yzW8TAAAAZAAE2s4AAAABAAAAAAAAAAAAAAABAAAAAAAAAAEAAAAAoUKq+1Z2GGB98qurLSmocHafvG6S+YzKNE6oiHIXo6kAAAABVVNEAAAAAACnUE2lfwuFZ+G+dkc+qiL0MwxB0CoR0au324j+JC9exQAAAAAdzWUAAAAAAAAAAAA=",
    }
    process_withdrawal(mock_response, transaction)
    assert (
        Transaction.objects.get(id=transaction_id).status
        == Transaction.STATUS.pending_stellar
    )


@pytest.mark.django_db
@patch("polaris.helpers.check_auth", side_effect=mock_check_auth_success)
def test_withdraw_interactive_success_transaction_successful(
    mock_check, client, acc1_usd_withdrawal_transaction_factory
):
    """
    `GET /transactions/withdraw/webapp` changes transaction to `completed`
    with successful transaction.
    """
    del mock_check
    acc1_usd_withdrawal_transaction_factory()
    response = client.post(WITHDRAW_PATH, {"asset_code": "USD"}, follow=True)
    content = json.loads(response.content)
    assert content["type"] == "interactive_customer_info_needed"

    transaction_id = content["id"]
    url = content["url"]
    response = client.post(
        url, {"amount": 50, "bank_account": "123456", "bank": "Bank"}
    )
    assert response.status_code == 200
    transaction = Transaction.objects.get(id=transaction_id)
    assert transaction.status == Transaction.STATUS.pending_user_transfer_start

    withdraw_memo = transaction.withdraw_memo
    mock_response = {
        "memo_type": "hash",
        "memo": format_memo_horizon(withdraw_memo),
        "successful": True,
        "id": "c5e8ada72c0e3c248ac7e1ec0ec97e204c06c295113eedbe632020cd6dc29ff8",
        "envelope_xdr": "AAAAAEU1B1qeJrucdqkbk1mJsnuFaNORfrOAzJyaAy1yzW8TAAAAZAAE2s4AAAABAAAAAAAAAAAAAAABAAAAAAAAAAEAAAAAoUKq+1Z2GGB98qurLSmocHafvG6S+YzKNE6oiHIXo6kAAAABVVNEAAAAAACnUE2lfwuFZ+G+dkc+qiL0MwxB0CoR0au324j+JC9exQAAAAAdzWUAAAAAAAAAAAA=",
    }
    process_withdrawal(mock_response, transaction)

    assert transaction.status == Transaction.STATUS.completed
    assert transaction.completed_at


@pytest.mark.django_db
def test_withdraw_authenticated_success(
    client, acc1_usd_withdrawal_transaction_factory
):
    """`GET /withdraw` succeeds with the SEP 10 authentication flow."""
    client_address = "GDKFNRUATPH4BSZGVFDRBIGZ5QAFILVFRIRYNSQ4UO7V2ZQAPRNL73RI"
    client_seed = "SDKWSBERDHP3SXW5A3LXSI7FWMMO5H7HG33KNYBKWH2HYOXJG2DXQHQY"
    acc1_usd_withdrawal_transaction_factory()

    # SEP 10.
    response = client.get(f"/auth?account={client_address}", follow=True)
    content = json.loads(response.content)

    envelope_xdr = content["transaction"]
    envelope_object = TransactionEnvelope.from_xdr(envelope_xdr, network_passphrase=settings.STELLAR_NETWORK_PASSPHRASE)
    client_signing_key = Keypair.from_secret(client_seed)
    envelope_object.sign(client_signing_key)
    client_signed_envelope_xdr = envelope_object.to_xdr()

    response = client.post(
        "/auth",
        data={"transaction": client_signed_envelope_xdr},
        content_type="application/json",
    )
    content = json.loads(response.content)
    encoded_jwt = content["token"]
    assert encoded_jwt

    header = {"HTTP_AUTHORIZATION": f"Bearer {encoded_jwt}"}
    response = client.post(WITHDRAW_PATH, {"asset_code": "USD"}, follow=True, **header)
    content = json.loads(response.content)
    assert content["type"] == "interactive_customer_info_needed"


@pytest.mark.django_db
def test_withdraw_no_jwt(client, acc1_usd_withdrawal_transaction_factory):
    """`GET /withdraw` fails if a required JWT isn't provided."""
    acc1_usd_withdrawal_transaction_factory()
    response = client.post(WITHDRAW_PATH, {"asset_code": "USD"}, follow=True)
    content = json.loads(response.content)
    assert response.status_code == 400
    assert content == {"error": "JWT must be passed as 'Authorization' header", "status_code": 400}



"""This module tests the `/transactions` endpoint."""
import json
import urllib
from unittest.mock import patch

import pytest
from django.conf import settings
from stellar_sdk.keypair import Keypair
from stellar_sdk.transaction_envelope import TransactionEnvelope

from polaris.tests.helpers import mock_check_auth_success


@pytest.mark.django_db
@patch("polaris.helpers.check_auth", side_effect=mock_check_auth_success)
def test_required_fields(mock_check, client, acc2_eth_withdrawal_transaction_factory):
    """Fails without required parameters."""
    del mock_check
    acc2_eth_withdrawal_transaction_factory()

    response = client.get(f"/transactions", follow=True)

    content = json.loads(response.content)
    assert response.status_code == 400
    assert content.get("error") is not None


@pytest.mark.django_db
@patch("polaris.helpers.check_auth", side_effect=mock_check_auth_success)
def test_required_account(mock_check, client, acc2_eth_withdrawal_transaction_factory):
    """Fails without `account` parameter."""
    del mock_check
    withdrawal = acc2_eth_withdrawal_transaction_factory()

    response = client.get(
        f"/transactions?asset_code={withdrawal.asset.code}", follow=True
    )

    content = json.loads(response.content)
    assert response.status_code == 400
    assert content.get("error") is not None


@pytest.mark.django_db
@patch("polaris.helpers.check_auth", side_effect=mock_check_auth_success)
def test_required_asset_code(
    mock_check, client, acc2_eth_withdrawal_transaction_factory
):
    """Fails without `asset_code` parameter."""
    del mock_check
    withdrawal = acc2_eth_withdrawal_transaction_factory()

    response = client.get(
        f"/transactions?account={withdrawal.stellar_account}", follow=True
    )

    content = json.loads(response.content)
    assert response.status_code == 400
    assert content.get("error") is not None


@pytest.mark.django_db
@patch("polaris.helpers.check_auth", side_effect=mock_check_auth_success)
def test_transactions_format(
    mock_check,
    client,
    acc2_eth_withdrawal_transaction_factory,
    acc2_eth_deposit_transaction_factory,
):
    """Response has correct length and status code."""
    del mock_check
    withdrawal = acc2_eth_withdrawal_transaction_factory()
    acc2_eth_deposit_transaction_factory()

    response = client.get(
        f"/transactions?asset_code={withdrawal.asset.code}&account={withdrawal.stellar_account}",
        follow=True,
    )
    content = json.loads(response.content)

    assert len(content.get("transactions")) == 2
    assert response.status_code == 200


@pytest.mark.django_db
@patch("polaris.helpers.check_auth", side_effect=mock_check_auth_success)
def test_transactions_order(
    mock_check,
    client,
    acc2_eth_withdrawal_transaction_factory,
    acc2_eth_deposit_transaction_factory,
):
    """Transactions are serialized in expected order."""
    del mock_check
    acc2_eth_deposit_transaction_factory()  # older transaction
    withdrawal = acc2_eth_withdrawal_transaction_factory()  # newer transaction

    response = client.get(
        f"/transactions?asset_code={withdrawal.asset.code}&account={withdrawal.stellar_account}",
        follow=True,
    )
    content = json.loads(response.content)

    # Withdrawal comes first, since transactions are ordered by -id
    withdrawal_transaction = content.get("transactions")[0]
    deposit_transaction = content.get("transactions")[1]

    assert withdrawal_transaction["kind"] == "withdrawal"
    assert deposit_transaction["kind"] == "deposit"


@pytest.mark.django_db
@patch("polaris.helpers.check_auth", side_effect=mock_check_auth_success)
def test_transactions_content(
    mock_check,
    client,
    acc2_eth_deposit_transaction_factory,
    acc2_eth_withdrawal_transaction_factory,
):
    """
    This expected response was adapted from the example SEP-0024 response on
    https://github.com/stellar/stellar-protocol/blob/master/ecosystem/sep-0024.md#transaction-history
    Some changes have been applied, to ensure the data we provide is in a consistent format and
    in accordance with design decisions from this reference implementation:

    - amounts are floats, so values like "500" are displayed as "500.0"
    - nullable fields are displayed, but with a null value
    """
    del mock_check
    deposit = acc2_eth_deposit_transaction_factory()
    withdrawal = acc2_eth_withdrawal_transaction_factory()

    d_started_at = deposit.started_at.isoformat().replace("+00:00", "Z")
    w_started_at = withdrawal.started_at.isoformat().replace("+00:00", "Z")
    w_completed_at = withdrawal.completed_at.isoformat().replace("+00:00", "Z")

    response = client.get(
        f"/transactions?asset_code={withdrawal.asset.code}&account={withdrawal.stellar_account}",
        follow=True,
    )
    content = json.loads(response.content)

    withdrawal_transaction = content.get("transactions")[0]
    deposit_transaction = content.get("transactions")[1]

    # Verifying the withdrawal transaction data:
    assert withdrawal_transaction["id"] == str(withdrawal.id)
    assert withdrawal_transaction["kind"] == withdrawal.kind
    assert withdrawal_transaction["status"] == withdrawal.status
    assert withdrawal_transaction["status_eta"] == 3600
    assert withdrawal_transaction["amount_in"] == str(withdrawal.amount_in)
    assert withdrawal_transaction["amount_out"] == str(withdrawal.amount_out)
    assert withdrawal_transaction["amount_fee"] == str(float(withdrawal.amount_fee))
    assert withdrawal_transaction["started_at"] == w_started_at
    assert withdrawal_transaction["completed_at"] == w_completed_at
    assert (
        withdrawal_transaction["stellar_transaction_id"]
        == withdrawal.stellar_transaction_id
    )
    assert (
        withdrawal_transaction["external_transaction_id"]
        == withdrawal.external_transaction_id
    )
    assert withdrawal_transaction["from_address"] is None
    assert withdrawal_transaction["to_address"] is None
    assert withdrawal_transaction["external_extra"] is None
    assert withdrawal_transaction["external_extra_text"] is None
    assert withdrawal_transaction["deposit_memo"] is None
    assert withdrawal_transaction["deposit_memo_type"] == withdrawal.deposit_memo_type
    assert (
        withdrawal_transaction["withdraw_anchor_account"]
        == withdrawal.withdraw_anchor_account
    )
    assert withdrawal_transaction["withdraw_memo"] == withdrawal.withdraw_memo
    assert withdrawal_transaction["withdraw_memo_type"] == withdrawal.withdraw_memo_type

    # Verifying the deposit transaction data:
    assert deposit_transaction["id"] == str(deposit.id)
    assert deposit_transaction["kind"] == deposit.kind
    assert deposit_transaction["status"] == deposit.status
    assert deposit_transaction["status_eta"] == deposit.status_eta
    assert deposit_transaction["amount_in"] == str(deposit.amount_in)
    assert deposit_transaction["amount_out"] == str(deposit.amount_out)
    assert deposit_transaction["amount_fee"] == str(float(deposit.amount_fee))
    assert deposit_transaction["started_at"] == d_started_at
    assert deposit_transaction["completed_at"] is None
    assert deposit_transaction["stellar_transaction_id"] is None
    assert (
        deposit_transaction["external_transaction_id"]
        == deposit.external_transaction_id
    )
    assert deposit_transaction["from_address"] is None
    assert deposit_transaction["to_address"] is None
    assert deposit_transaction["external_extra"] is None
    assert deposit_transaction["external_extra_text"] is None
    assert deposit_transaction["deposit_memo"] == deposit.deposit_memo
    assert deposit_transaction["deposit_memo_type"] == deposit.deposit_memo_type
    assert deposit_transaction["withdraw_anchor_account"] is None
    assert deposit_transaction["withdraw_memo"] is None
    assert deposit_transaction["withdraw_memo_type"] == deposit.withdraw_memo_type

    # stellar_account and asset should not be exposed:
    with pytest.raises(KeyError):
        assert withdrawal_transaction["stellar_account"]
    with pytest.raises(KeyError):
        assert withdrawal_transaction["asset"]
    with pytest.raises(KeyError):
        assert deposit_transaction["stellar_account"]
    with pytest.raises(KeyError):
        assert deposit_transaction["asset"]


@pytest.mark.django_db
@patch("polaris.helpers.check_auth", side_effect=mock_check_auth_success)
def test_paging_id(
    mock_check,
    client,
    acc2_eth_deposit_transaction_factory,
    acc2_eth_withdrawal_transaction_factory,
):
    """Only return transactions chronologically after a `paging_id`, if provided."""
    del mock_check
    acc2_eth_deposit_transaction_factory()
    withdrawal = acc2_eth_withdrawal_transaction_factory()

    response = client.get(
        (
            f"/transactions?asset_code={withdrawal.asset.code}"
            f"&account={withdrawal.stellar_account}"
            f"&paging_id={withdrawal.id}"
        ),
        follow=True,
    )
    content = json.loads(response.content)

    # By providing the paging_id = w.id, we're looking for entries older than `w`
    # which only leaves us with the deposit transaction.
    assert len(content.get("transactions")) == 1
    assert content.get("transactions")[0]["kind"] == "deposit"


@pytest.mark.django_db
@patch("polaris.helpers.check_auth", side_effect=mock_check_auth_success)
def test_kind_filter(
    mock_check,
    client,
    acc2_eth_deposit_transaction_factory,
    acc2_eth_withdrawal_transaction_factory,
):
    """Valid `kind` succeeds."""
    del mock_check
    acc2_eth_deposit_transaction_factory()
    withdrawal = acc2_eth_withdrawal_transaction_factory()

    response = client.get(
        (
            f"/transactions?asset_code={withdrawal.asset.code}"
            f"&account={withdrawal.stellar_account}"
            f"&kind=deposit"
        ),
        follow=True,
    )
    content = json.loads(response.content)

    # By providing the paging_id = w.id, we're looking for entries older than `w`
    # which only leaves us with the deposit transaction.
    assert len(content.get("transactions")) == 1
    assert content.get("transactions")[0]["kind"] == "deposit"


@pytest.mark.django_db
@patch("polaris.helpers.check_auth", side_effect=mock_check_auth_success)
def test_kind_filter_no_500(
    mock_check,
    client,
    acc2_eth_deposit_transaction_factory,
    acc2_eth_withdrawal_transaction_factory,
):
    """Invalid `kind` fails gracefully."""
    del mock_check
    acc2_eth_deposit_transaction_factory()
    withdrawal = acc2_eth_withdrawal_transaction_factory()

    response = client.get(
        (
            f"/transactions?asset_code={withdrawal.asset.code}"
            f"&account={withdrawal.stellar_account}&kind=somethingelse"
        ),
        follow=True,
    )
    content = json.loads(response.content)

    # By providing the paging_id = w.id, we're looking for entries older than `w`
    # which only leaves us with the deposit transaction.
    assert not content.get("transactions")


@pytest.mark.django_db
@patch("polaris.helpers.check_auth", side_effect=mock_check_auth_success)
def test_limit(
    mock_check,
    client,
    acc2_eth_deposit_transaction_factory,
    acc2_eth_withdrawal_transaction_factory,
):
    """Valid `limit` succeeds."""
    del mock_check
    acc2_eth_deposit_transaction_factory()
    withdrawal = acc2_eth_withdrawal_transaction_factory()  # newest

    response = client.get(
        f"/transactions?asset_code={withdrawal.asset.code}"
        f"&account={withdrawal.stellar_account}&limit=1",
        follow=True,
    )
    content = json.loads(response.content)

    # By providing the paging_id = w.id, we're looking for entries older than `w`
    # which only leaves us with the deposit transaction.
    assert len(content.get("transactions")) == 1
    assert content.get("transactions")[0]["kind"] == "withdrawal"


@pytest.mark.django_db
@patch("polaris.helpers.check_auth", side_effect=mock_check_auth_success)
def test_invalid_limit(
    mock_check,
    client,
    acc2_eth_deposit_transaction_factory,
    acc2_eth_withdrawal_transaction_factory,
):
    """Non-integer `limit` fails."""
    del mock_check
    acc2_eth_deposit_transaction_factory()
    withdrawal = acc2_eth_withdrawal_transaction_factory()  # newest

    response = client.get(
        (
            f"/transactions?asset_code={withdrawal.asset.code}"
            f"&account={withdrawal.stellar_account}&limit=string"
        ),
        follow=True,
    )
    content = json.loads(response.content)

    assert content.get("error") is not None
    assert response.status_code == 400


@pytest.mark.django_db
@patch("polaris.helpers.check_auth", side_effect=mock_check_auth_success)
def test_negative_limit(
    mock_check,
    client,
    acc2_eth_deposit_transaction_factory,
    acc2_eth_withdrawal_transaction_factory,
):
    """Negative `limit` fails."""
    del mock_check
    acc2_eth_deposit_transaction_factory()
    withdrawal = acc2_eth_withdrawal_transaction_factory()  # newest

    response = client.get(
        (
            f"/transactions?asset_code={withdrawal.asset.code}"
            f"&account={withdrawal.stellar_account}&limit=-1"
        ),
        follow=True,
    )
    content = json.loads(response.content)

    assert content.get("error") is not None
    assert response.status_code == 400


@pytest.mark.django_db
@patch("polaris.helpers.check_auth", side_effect=mock_check_auth_success)
def test_no_older_than_filter(
    mock_check,
    client,
    acc2_eth_deposit_transaction_factory,
    acc2_eth_withdrawal_transaction_factory,
):
    """Valid `no_older_than` succeeds."""
    del mock_check
    withdrawal_transaction = (
        acc2_eth_withdrawal_transaction_factory()
    )  # older transaction
    deposit_transaction = acc2_eth_deposit_transaction_factory()  # newer transaction

    urlencoded_datetime = urllib.parse.quote(deposit_transaction.started_at.isoformat())
    response = client.get(
        (
            f"/transactions?asset_code={withdrawal_transaction.asset.code}"
            f"&account={withdrawal_transaction.stellar_account}"
            f"&no_older_than={urlencoded_datetime}"
        ),
        follow=True,
    )
    content = json.loads(response.content)

    assert response.status_code == 200
    assert len(content.get("transactions")) == 1
    assert content.get("transactions")[0]["kind"] == "deposit"


@pytest.mark.django_db
def test_transactions_authenticated_success(
    client,
    acc2_eth_withdrawal_transaction_factory,
    acc2_eth_deposit_transaction_factory,
):
    """
    Response has correct length and status code, if the SEP 10 authentication
    token is required.
    """
    client_address = "GDKFNRUATPH4BSZGVFDRBIGZ5QAFILVFRIRYNSQ4UO7V2ZQAPRNL73RI"
    client_seed = "SDKWSBERDHP3SXW5A3LXSI7FWMMO5H7HG33KNYBKWH2HYOXJG2DXQHQY"
    withdrawal = acc2_eth_withdrawal_transaction_factory()
    withdrawal.stellar_address = client_address
    withdrawal.save()
    acc2_eth_deposit_transaction_factory()

    # SEP 10.
    response = client.get(f"/auth?account={client_address}", follow=True)
    content = json.loads(response.content)
    envelope_xdr = content["transaction"]
    envelope_object = TransactionEnvelope.from_xdr(envelope_xdr, network_passphrase=settings.STELLAR_NETWORK_PASSPHRASE)
    client_signing_key = Keypair.from_secret(client_seed)
    envelope_object.sign(client_signing_key)
    client_signed_envelope_xdr = envelope_object.to_xdr()

    response = client.post(
        "/auth",
        data={"transaction": client_signed_envelope_xdr},
        content_type="application/json",
    )
    content = json.loads(response.content)
    encoded_jwt = content["token"]
    assert encoded_jwt

    # For testing, we make the key `HTTP_AUTHORIZATION`. This is the value that
    # we expect due to the middleware.
    header = {"HTTP_AUTHORIZATION": f"Bearer {encoded_jwt}"}

    response = client.get(
        f"/transactions?asset_code={withdrawal.asset.code}&account={withdrawal.stellar_account}",
        follow=True,
        **header,
    )
    content = json.loads(response.content)

    assert len(content.get("transactions")) == 2
    assert response.status_code == 200


@pytest.mark.django_db
def test_transactions_no_jwt(
    client, acc2_eth_withdrawal_transaction_factory
):
    """`GET /transactions` fails if a required JWT is not provided."""
    withdrawal = acc2_eth_withdrawal_transaction_factory()
    response = client.get(
        f"/transactions?asset_code={withdrawal.asset.code}&account={withdrawal.stellar_account}",
        follow=True,
    )
    content = json.loads(response.content)
    assert response.status_code == 400
    assert content == {"error": "JWT must be passed as 'Authorization' header", "status_code": 400}



"""This module tests `GET /transaction`."""
import json
from unittest.mock import patch

import pytest
from stellar_sdk.keypair import Keypair
from stellar_sdk.transaction_envelope import TransactionEnvelope

from polaris import settings
from polaris.models import Transaction
from polaris.tests.helpers import mock_check_auth_success


@pytest.mark.django_db
@patch("polaris.helpers.check_auth", side_effect=mock_check_auth_success)
def test_transaction_requires_param(
    mock_check,
    client,
    acc1_usd_deposit_transaction_factory,
    acc2_eth_withdrawal_transaction_factory,
):
    """Fails without parameters."""
    del mock_check
    acc1_usd_deposit_transaction_factory()
    acc2_eth_withdrawal_transaction_factory()

    response = client.get(f"/transaction", follow=True)
    content = json.loads(response.content)

    assert response.status_code == 400
    assert content.get("error")


@pytest.mark.django_db
@patch("polaris.helpers.check_auth", side_effect=mock_check_auth_success)
def test_transaction_id_filter_and_format(
    mock_check,
    client,
    acc1_usd_deposit_transaction_factory,
    acc2_eth_withdrawal_transaction_factory,
):
    """Succeeds with expected response if `id` provided."""
    del mock_check
    acc1_usd_deposit_transaction_factory()
    withdrawal = acc2_eth_withdrawal_transaction_factory()

    w_started_at = withdrawal.started_at.isoformat().replace("+00:00", "Z")
    w_completed_at = withdrawal.completed_at.isoformat().replace("+00:00", "Z")

    response = client.get(f"/transaction?id={withdrawal.id}", follow=True)
    content = json.loads(response.content)

    assert response.status_code == 200

    withdrawal_transaction = content.get("transaction")

    # Verifying the withdrawal transaction data:
    assert isinstance(withdrawal_transaction["id"], str)
    assert withdrawal_transaction["kind"] == "withdrawal"
    assert withdrawal_transaction["status"] == "completed"
    assert withdrawal_transaction["status_eta"] == 3600
    assert withdrawal_transaction["amount_in"] == "500.0"
    assert withdrawal_transaction["amount_out"] == "495.0"
    assert withdrawal_transaction["amount_fee"] == "3.0"
    assert withdrawal_transaction["started_at"] == w_started_at
    assert withdrawal_transaction["completed_at"] == w_completed_at
    assert (
        withdrawal_transaction["stellar_transaction_id"]
        == "17a670bc424ff5ce3b386dbfaae9990b66a2a37b4fbe51547e8794962a3f9e6a"
    )
    assert (
        withdrawal_transaction["external_transaction_id"]
        == "2dd16cb409513026fbe7defc0c6f826c2d2c65c3da993f747d09bf7dafd31094"
    )
    assert withdrawal_transaction["from_address"] is None
    assert withdrawal_transaction["to_address"] is None
    assert withdrawal_transaction["external_extra"] is None
    assert withdrawal_transaction["external_extra_text"] is None
    assert withdrawal_transaction["deposit_memo"] is None
    assert withdrawal_transaction["deposit_memo_type"] == withdrawal.deposit_memo_type
    assert (
        withdrawal_transaction["withdraw_anchor_account"]
        == withdrawal.withdraw_anchor_account
    )
    assert withdrawal_transaction["withdraw_memo"] == withdrawal.withdraw_memo
    assert withdrawal_transaction["withdraw_memo_type"] == withdrawal.withdraw_memo_type


@pytest.mark.django_db
@patch("polaris.helpers.check_auth", side_effect=mock_check_auth_success)
def test_transaction_stellar_filter(
    mock_check,
    client,
    acc1_usd_deposit_transaction_factory,
    acc2_eth_withdrawal_transaction_factory,
):
    """Succeeds with expected response if `stellar_transaction_id` provided."""
    del mock_check
    acc1_usd_deposit_transaction_factory()
    withdrawal = acc2_eth_withdrawal_transaction_factory()

    response = client.get(
        f"/transaction?stellar_transaction_id={withdrawal.stellar_transaction_id}",
        follow=True,
    )
    content = json.loads(response.content)

    assert response.status_code == 200

    withdrawal_transaction = content.get("transaction")
    assert withdrawal_transaction["kind"] == "withdrawal"
    assert withdrawal_transaction["status"] == Transaction.STATUS.completed


@pytest.mark.django_db
@patch("polaris.helpers.check_auth", side_effect=mock_check_auth_success)
def test_transaction_external_filter(
    mock_check,
    client,
    acc1_usd_deposit_transaction_factory,
    acc2_eth_withdrawal_transaction_factory,
):
    """Succeeds with expected response if `external_transaction_id` provided."""
    del mock_check
    deposit = acc1_usd_deposit_transaction_factory()
    acc2_eth_withdrawal_transaction_factory()

    response = client.get(
        f"/transaction?external_transaction_id={deposit.external_transaction_id}",
        follow=True,
    )
    content = json.loads(response.content)

    assert response.status_code == 200

    withdrawal_transaction = content.get("transaction")
    assert withdrawal_transaction["kind"] == "deposit"
    assert (
        withdrawal_transaction["status"]
        == Transaction.STATUS.pending_user_transfer_start
    )


@pytest.mark.django_db
@patch("polaris.helpers.check_auth", side_effect=mock_check_auth_success)
def test_transaction_multiple_filters(
    mock_check,
    client,
    acc1_usd_deposit_transaction_factory,
    acc2_eth_withdrawal_transaction_factory,
):
    """Succeeds with expected response if multiple valid ids provided."""
    del mock_check
    acc1_usd_deposit_transaction_factory()
    withdrawal = acc2_eth_withdrawal_transaction_factory()

    response = client.get(
        (
            f"/transaction?id={withdrawal.id}"
            f"&external_transaction_id={withdrawal.external_transaction_id}"
            f"&stellar_transaction_id={withdrawal.stellar_transaction_id}"
        ),
        follow=True,
    )
    content = json.loads(response.content)

    assert response.status_code == 200

    withdrawal_transaction = content.get("transaction")

    # Verifying the withdrawal transaction data:
    assert isinstance(withdrawal_transaction["id"], str)
    assert withdrawal_transaction["kind"] == "withdrawal"
    assert withdrawal_transaction["status"] == Transaction.STATUS.completed


@pytest.mark.django_db
@patch("polaris.helpers.check_auth", side_effect=mock_check_auth_success)
def test_transaction_filtering_no_result(
    mock_check,
    client,
    acc1_usd_deposit_transaction_factory,
    acc2_eth_withdrawal_transaction_factory,
):
    """Succeeds with expected response if invalid combo of ids provided."""
    del mock_check
    deposit = acc1_usd_deposit_transaction_factory()
    withdrawal = acc2_eth_withdrawal_transaction_factory()

    response = client.get(
        (
            f"/transaction?id={deposit.id}"
            f"&external_transaction_id={withdrawal.external_transaction_id}"
            f"&stellar_transaction_id={withdrawal.stellar_transaction_id}"
        ),
        follow=True,
    )
    content = json.loads(response.content)

    assert response.status_code == 404
    assert content.get("error") is not None


@pytest.mark.django_db
def test_transaction_authenticated_success(
    client,
    acc1_usd_deposit_transaction_factory,
    acc2_eth_withdrawal_transaction_factory,
):
    """
    Succeeds with expected response if authentication required.
    Though it filters using the stellar transaction ID, the logic
    should apply in any case.
    """
    client_address = "GDKFNRUATPH4BSZGVFDRBIGZ5QAFILVFRIRYNSQ4UO7V2ZQAPRNL73RI"
    client_seed = "SDKWSBERDHP3SXW5A3LXSI7FWMMO5H7HG33KNYBKWH2HYOXJG2DXQHQY"
    acc1_usd_deposit_transaction_factory()
    withdrawal = acc2_eth_withdrawal_transaction_factory()
    withdrawal.stellar_address = client_address
    withdrawal.save()

    # SEP 10.
    response = client.get(f"/auth?account={client_address}", follow=True)
    content = json.loads(response.content)
    envelope_xdr = content["transaction"]
    envelope_object = TransactionEnvelope.from_xdr(envelope_xdr, network_passphrase=settings.STELLAR_NETWORK_PASSPHRASE)
    client_signing_key = Keypair.from_secret(client_seed)
    envelope_object.sign(client_signing_key)
    client_signed_envelope_xdr = envelope_object.to_xdr()

    response = client.post(
        "/auth",
        data={"transaction": client_signed_envelope_xdr},
        content_type="application/json",
    )
    content = json.loads(response.content)
    encoded_jwt = content["token"]
    assert encoded_jwt

    # For testing, we make the key `HTTP_AUTHORIZATION`. This is the value that
    # we expect due to the middleware.
    header = {"HTTP_AUTHORIZATION": f"Bearer {encoded_jwt}"}
    response = client.get(
        f"/transaction?stellar_transaction_id={withdrawal.stellar_transaction_id}",
        follow=True,
        **header,
    )
    content = json.loads(response.content)

    assert response.status_code == 200

    withdrawal_transaction = content.get("transaction")
    assert withdrawal_transaction["kind"] == "withdrawal"
    assert withdrawal_transaction["status"] == Transaction.STATUS.completed


@pytest.mark.django_db
def test_transaction_no_jwt(
    client, acc2_eth_withdrawal_transaction_factory
):
    """Fails if required JWT is not provided."""
    withdrawal = acc2_eth_withdrawal_transaction_factory()

    response = client.get(
        (
            f"/transaction?id={withdrawal.id}"
            f"&external_transaction_id={withdrawal.external_transaction_id}"
            f"&stellar_transaction_id={withdrawal.stellar_transaction_id}"
        ),
        follow=True,
    )
    content = json.loads(response.content)

    assert response.status_code == 400
    assert content == {"error": "JWT must be passed as 'Authorization' header", "status_code": 400}



"""
This module implements the logic for the `/.well-known` endpoint.
In particular, this generates the `.toml` file for the anchor server.

The importance of the Stellar TOML file is explained in SEP-0001:
https://github.com/stellar/stellar-protocol/blob/master/ecosystem/sep-0001.md
"""

import toml
from django.http.response import HttpResponse
from polaris import settings
from rest_framework.decorators import api_view
from polaris.models import Asset


def generate_toml(request):
    """Generate the TOML file."""
    toml_dict = {}

    # Globals.
    toml_dict["TRANSFER_SERVER"] = request.build_absolute_uri("/")
    toml_dict["WEB_AUTH_ENDPOINT"] = request.build_absolute_uri("/auth")
    toml_dict["ACCOUNTS"] = [settings.STELLAR_DISTRIBUTION_ACCOUNT_ADDRESS]
    toml_dict["VERSION"] = "0.1.0"

    toml_dict["DOCUMENTATION"] = {
        "ORG_NAME": "Stellar Development Foundation",
        "ORG_URL": "https://stellar.org",
        "ORG_DESCRIPTION": "SEP 24 reference server.",
        "ORG_KEYBASE": "stellar.public",
        "ORG_TWITTER": "StellarOrg",
        "ORG_GITHUB": "stellar",
    }

    toml_dict["CURRENCIES"] = [
        {"code": asset.code, "issuer": settings.STELLAR_ISSUER_ACCOUNT_ADDRESS}
        for asset in Asset.objects.all().iterator()
    ]

    return HttpResponse(toml.dumps(toml_dict), content_type="text/plain")



"""This module defines a serializer for the transaction model."""
from rest_framework import serializers

from polaris.models import Transaction


class TransactionSerializer(serializers.ModelSerializer):
    """Defines the custom serializer for a transaction."""

    id = serializers.CharField()
    amount_in = serializers.CharField()
    amount_out = serializers.CharField()
    amount_fee = serializers.CharField()
    more_info_url = serializers.SerializerMethodField()

    def get_more_info_url(self, transaction_instance):
        return self.context.get("more_info_url")

    class Meta:
        model = Transaction
        fields = [
            "id",
            "kind",
            "status",
            "status_eta",
            "amount_in",
            "amount_out",
            "amount_fee",
            "started_at",
            "completed_at",
            "stellar_transaction_id",
            "external_transaction_id",
            "from_address",
            "to_address",
            "external_extra",
            "external_extra_text",
            "deposit_memo",
            "deposit_memo_type",
            "withdraw_anchor_account",
            "withdraw_memo",
            "withdraw_memo_type",
            "more_info_url",
        ]



"""
This module tests the `/deposit` endpoint.
Celery tasks are called synchronously. Horizon calls are mocked for speed and correctness.
"""
import json
import uuid
from unittest.mock import patch

import pytest
from stellar_sdk import Keypair, TransactionEnvelope
from stellar_sdk.client.response import Response
from stellar_sdk.exceptions import BadRequestError
from django.core.management import call_command
    
from polaris import settings
from polaris.management.commands.create_stellar_deposit import (
    SUCCESS_XDR,
    TRUSTLINE_FAILURE_XDR,
)
from polaris.models import Transaction
from polaris.tests.helpers import mock_check_auth_success, mock_load_not_exist_account

DEPOSIT_PATH = f"/transactions/deposit/interactive"
HORIZON_SUCCESS_RESPONSE = {"result_xdr": SUCCESS_XDR, "hash": "test_stellar_id"}

@pytest.mark.django_db
@patch("polaris.helpers.check_auth", side_effect=mock_check_auth_success)
def test_deposit_success(mock_check, client, acc1_usd_deposit_transaction_factory):
    """`POST /transactions/deposit/interactive` succeeds with no optional arguments."""
    del mock_check
    deposit = acc1_usd_deposit_transaction_factory()
    response = client.post(
        DEPOSIT_PATH, {"asset_code": "USD", "account": deposit.stellar_account},
    )
    content = json.loads(response.content)
    assert content["type"] == "interactive_customer_info_needed"


@pytest.mark.django_db
@patch("polaris.helpers.check_auth", side_effect=mock_check_auth_success)
def test_deposit_success_memo(mock_check, client, acc1_usd_deposit_transaction_factory):
    """`POST /transactions/deposit/interactive` succeeds with valid `memo` and `memo_type`."""
    del mock_check
    deposit = acc1_usd_deposit_transaction_factory()
    response = client.post(
        DEPOSIT_PATH, {"asset_code": "USD", "account": deposit.stellar_account, "memo_type": "text", "memo": "foo"},
        follow=True,
    )

    content = json.loads(response.content)
    assert content["type"] == "interactive_customer_info_needed"


@patch("polaris.helpers.check_auth", side_effect=mock_check_auth_success)
def test_deposit_no_params(mock_check, client):
    """`POST /transactions/deposit/interactive` fails with no required parameters."""
    # Because this test does not use the database, the changed setting
    # earlier in the file is not persisted when the tests not requiring
    # a database are run. Thus, we set that flag again here.
    del mock_check
    response = client.post(
        DEPOSIT_PATH, {}, follow=True)
    content = json.loads(response.content)

    assert response.status_code == 400
    assert content == {"error": "`asset_code` and `account` are required parameters", "status_code": 400}

@patch("polaris.helpers.check_auth", side_effect=mock_check_auth_success)
def test_deposit_no_account(mock_check, client):
    """`POST /transactions/deposit/interactive` fails with no `account` parameter."""
    del mock_check
    response = client.post(DEPOSIT_PATH, {"asset_code": "NADA"}, follow=True)
    content = json.loads(response.content)

    assert response.status_code == 400
    assert content == {"error": "`asset_code` and `account` are required parameters", "status_code": 400}


@pytest.mark.django_db
@patch("polaris.helpers.check_auth", side_effect=mock_check_auth_success)
def test_deposit_no_asset(mock_check, client, acc1_usd_deposit_transaction_factory):
    """`POST /transactions/deposit/interactive` fails with no `asset_code` parameter."""
    del mock_check
    deposit = acc1_usd_deposit_transaction_factory()
    response = client.post(DEPOSIT_PATH, {"account": deposit.stellar_account}, follow=True)
    content = json.loads(response.content)

    assert response.status_code == 400
    assert content == {"error": "`asset_code` and `account` are required parameters", "status_code": 400}


@pytest.mark.django_db
@patch("polaris.helpers.check_auth", side_effect=mock_check_auth_success)
def test_deposit_invalid_account(
    mock_check, client, acc1_usd_deposit_transaction_factory
):
    """`POST /transactions/deposit/interactive` fails with an invalid `account` parameter."""
    del mock_check
    acc1_usd_deposit_transaction_factory()
    response = client.post(
        DEPOSIT_PATH, {"asset_code": "USD", "account": "GBSH7WNSDU5FEIED2JQZIOQPZXREO3YNH2M5DIBE8L2X5OOAGZ7N2QI6"},
        follow=True,
    )
    content = json.loads(response.content)

    assert response.status_code == 400
    assert content == {"error": "invalid 'account'", "status_code": 400}


@pytest.mark.django_db
@patch("polaris.helpers.check_auth", side_effect=mock_check_auth_success)
def test_deposit_invalid_asset(
    mock_check, client, acc1_usd_deposit_transaction_factory
):
    """`POST /transactions/deposit/interactive` fails with an invalid `asset_code` parameter."""
    del mock_check
    deposit = acc1_usd_deposit_transaction_factory()
    response = client.post(
        DEPOSIT_PATH, {"asset_code": "GBP", "account": deposit.stellar_account}, follow=True
    )
    content = json.loads(response.content)

    assert response.status_code == 400
    assert content == {"error": "invalid operation for asset GBP", "status_code": 400}


@pytest.mark.django_db
@patch("polaris.helpers.check_auth", side_effect=mock_check_auth_success)
def test_deposit_invalid_memo_type(
    mock_check, client, acc1_usd_deposit_transaction_factory
):
    """`POST /transactions/deposit/interactive` fails with an invalid `memo_type` optional parameter."""
    del mock_check
    deposit = acc1_usd_deposit_transaction_factory()
    response = client.post(
        DEPOSIT_PATH, {"asset_code": "USD", "account": deposit.stellar_account, "memo_type": "test"},
        follow=True,
    )
    content = json.loads(response.content)

    assert response.status_code == 400
    assert content == {"error": "invalid 'memo_type'", "status_code": 400}


@pytest.mark.django_db
@patch("polaris.helpers.check_auth", side_effect=mock_check_auth_success)
def test_deposit_no_memo(mock_check, client, acc1_usd_deposit_transaction_factory):
    """`POST /transactions/deposit/interactive` fails with a valid `memo_type` and no `memo` provided."""
    del mock_check
    deposit = acc1_usd_deposit_transaction_factory()
    response = client.post(
        DEPOSIT_PATH, {"asset_code": "USD", "account": deposit.stellar_account, "memo_type": "text"},
        follow=True,
    )
    content = json.loads(response.content)
    print(response)
    assert response.status_code == 400
    assert content == {"error": "'memo_type' provided with no 'memo'", "status_code": 400}


@pytest.mark.django_db
@patch("polaris.helpers.check_auth", side_effect=mock_check_auth_success)
def test_deposit_no_memo_type(mock_check, client, acc1_usd_deposit_transaction_factory):
    """`POST /transactions/deposit/interactive` fails with a valid `memo` and no `memo_type` provided."""
    del mock_check
    deposit = acc1_usd_deposit_transaction_factory()
    response = client.post(
        DEPOSIT_PATH, {"asset_code": "USD", "account": deposit.stellar_account, "memo": "text"},
        follow=True,
    )
    content = json.loads(response.content)

    assert response.status_code == 400
    assert content == {"error": "'memo' provided with no 'memo_type'", "status_code": 400}


@pytest.mark.django_db
@patch("polaris.helpers.check_auth", side_effect=mock_check_auth_success)
def test_deposit_invalid_hash_memo(
    mock_check, client, acc1_usd_deposit_transaction_factory
):
    """`POST /transactions/deposit/interactive` fails with a valid `memo` of incorrect `memo_type` hash."""
    del mock_check
    deposit = acc1_usd_deposit_transaction_factory()
    response = client.post(
        DEPOSIT_PATH, {"asset_code": "USD", "account": deposit.stellar_account, "memo_type": "hash", "memo": "foo"},
        follow=True,
    )
    content = json.loads(response.content)

    assert response.status_code == 400
    assert content == {"error": "'memo' does not match memo_type' hash", "status_code": 400}


def test_deposit_confirm_no_txid(client):
    """`GET /transactions/deposit/confirm_transaction` fails with no `transaction_id`."""
    response = client.get(f"/transactions/deposit/confirm_transaction?amount=0", follow=True)
    content = json.loads(response.content)
    assert response.status_code == 400
    assert content == {
        "error": "no 'transaction_id' provided",
        "status_code": 400
    }


@pytest.mark.django_db
def test_deposit_confirm_invalid_txid(client):
    """`GET /transactions/deposit/confirm_transaction` fails with an invalid `transaction_id`."""
    incorrect_transaction_id = uuid.uuid4()
    response = client.get(
        f"/transactions/deposit/confirm_transaction?amount=0&transaction_id={incorrect_transaction_id}",
        follow=True,
    )
    content = json.loads(response.content)
    assert response.status_code == 400
    assert content == {
        "error": "no transaction with matching 'transaction_id' exists",
        "status_code": 400
    }


@pytest.mark.django_db
def test_deposit_confirm_no_amount(client, acc1_usd_deposit_transaction_factory):
    """`GET /transactions/deposit/confirm_transaction` fails with no `amount`."""
    deposit = acc1_usd_deposit_transaction_factory()
    response = client.get(
        f"/transactions/deposit/confirm_transaction?transaction_id={deposit.id}", follow=True
    )
    content = json.loads(response.content)
    assert response.status_code == 400
    assert content == {"error": "no 'amount' provided", "status_code": 400}


@pytest.mark.django_db
def test_deposit_confirm_invalid_amount(client, acc1_usd_deposit_transaction_factory):
    """`GET /transactions/deposit/confirm_transaction` fails with a non-float `amount`."""
    deposit = acc1_usd_deposit_transaction_factory()
    response = client.get(
        f"/transactions/deposit/confirm_transaction?transaction_id={deposit.id}&amount=foo",
        follow=True,
    )
    content = json.loads(response.content)
    assert response.status_code == 400
    assert content == {
        "error": "non-float 'amount' provided",
        "status_code": 400
    }


@pytest.mark.django_db
def test_deposit_confirm_incorrect_amount(client, acc1_usd_deposit_transaction_factory):
    """`GET /transactions/deposit/confirm_transaction` fails with an incorrect `amount`."""
    deposit = acc1_usd_deposit_transaction_factory()
    incorrect_amount = deposit.amount_in + 1
    response = client.get(
        f"/transactions/deposit/confirm_transaction?transaction_id={deposit.id}&amount={incorrect_amount}",
        follow=True,
    )
    content = json.loads(response.content)
    assert response.status_code == 400
    assert content == {
        "error": "incorrect 'amount' value for transaction with given 'transaction_id'",
        "status_code": 400
    }


@pytest.mark.django_db
@patch("stellar_sdk.server.Server.fetch_base_fee", return_value=100)
@patch("stellar_sdk.server.Server.submit_transaction", return_value=HORIZON_SUCCESS_RESPONSE)
def test_deposit_confirm_success(
    mock_submit,
    mock_base_fee,
    client,
    acc1_usd_deposit_transaction_factory,
):
    """`GET /transactions/deposit/confirm_transaction` succeeds with correct `amount` and `transaction_id`."""
    del mock_submit, mock_base_fee
    deposit = acc1_usd_deposit_transaction_factory()
    amount = deposit.amount_in
    response = client.get(
        f"/transactions/deposit/confirm_transaction?amount={amount}&transaction_id={deposit.id}",
        follow=True,
    )
    assert response.status_code == 200
    content = json.loads(response.content)
    transaction = content["transaction"]
    assert transaction
    assert transaction["status"] == Transaction.STATUS.pending_anchor
    assert float(transaction["amount_in"]) == amount
    assert int(transaction["status_eta"]) == 5


@pytest.mark.django_db
@patch("stellar_sdk.server.Server.fetch_base_fee", return_value=100)
@patch("stellar_sdk.server.Server.submit_transaction", return_value=HORIZON_SUCCESS_RESPONSE)
def test_deposit_confirm_external_id(
    mock_submit,
    mock_base_fee,
    client,
    acc1_usd_deposit_transaction_factory,
):
    """`GET /transactions/deposit/confirm_transaction` successfully stores an `external_id`."""
    del mock_submit, mock_base_fee
    deposit = acc1_usd_deposit_transaction_factory()
    amount = deposit.amount_in
    external_id = "foo"
    response = client.get(
        (
            f"/transactions/deposit/confirm_transaction?amount={amount}&transaction_id="
            f"{deposit.id}&external_transaction_id={external_id}"
        ),
        follow=True,
    )
    assert response.status_code == 200
    content = json.loads(response.content)
    transaction = content["transaction"]
    assert transaction
    assert transaction["status"] == Transaction.STATUS.pending_anchor
    assert float(transaction["amount_in"]) == amount
    assert int(transaction["status_eta"]) == 5
    assert transaction["external_transaction_id"] == external_id


@pytest.mark.django_db
@patch("stellar_sdk.server.Server.fetch_base_fee", return_value=100)
@patch(
    "stellar_sdk.server.Server.submit_transaction",
    side_effect=BadRequestError(
        response=Response(
            status_code=400,
            headers={},
            url="",
            text=json.dumps(dict(status=400, extras=dict(result_xdr=TRUSTLINE_FAILURE_XDR))),
        )
    ),
)
def test_deposit_stellar_no_trustline(
    mock_submit, mock_base_fee, client, acc1_usd_deposit_transaction_factory,
):
    """
    `create_stellar_deposit` sets the transaction with the provided `transaction_id` to
    status `pending_trust` if the provided transaction's Stellar account has no trustline
    for its asset. (We assume the asset's issuer is the server Stellar account.)
    """
    del mock_submit, mock_base_fee, client
    deposit = acc1_usd_deposit_transaction_factory()
    deposit.status = Transaction.STATUS.pending_anchor
    deposit.save()
    call_command("create_stellar_deposit", deposit.id)
    assert (
        Transaction.objects.get(id=deposit.id).status
        == Transaction.STATUS.pending_trust
    )


@pytest.mark.django_db
@patch("stellar_sdk.server.Server.fetch_base_fee", return_value=100)
@patch("stellar_sdk.server.Server.load_account", side_effect=mock_load_not_exist_account)
@patch("stellar_sdk.server.Server.submit_transaction", return_value=HORIZON_SUCCESS_RESPONSE,
)
def test_deposit_stellar_no_account(
    mock_load_account, mock_base_fee, client, acc1_usd_deposit_transaction_factory,
):
    """
    `create_stellar_deposit` sets the transaction with the provided `transaction_id` to
    status `pending_trust` if the provided transaction's `stellar_account` does not
    exist yet. This condition is mocked by throwing an error when attempting to load
    information for the provided account.
    Normally, this function creates the account. We have mocked out that functionality,
    as it relies on network calls to Horizon.
    """
    del mock_load_account, mock_base_fee, client
    deposit = acc1_usd_deposit_transaction_factory()
    deposit.status = Transaction.STATUS.pending_anchor
    deposit.save()
    call_command("create_stellar_deposit", deposit.id)
    assert (
        Transaction.objects.get(id=deposit.id).status
        == Transaction.STATUS.pending_trust
    )


@pytest.mark.django_db
@patch("stellar_sdk.server.Server.fetch_base_fee", return_value=100)
@patch("stellar_sdk.server.Server.submit_transaction", return_value=HORIZON_SUCCESS_RESPONSE)
def test_deposit_stellar_success(
    mock_submit, mock_base_fee, client, acc1_usd_deposit_transaction_factory,
):
    """
    `create_stellar_deposit` succeeds if the provided transaction's `stellar_account`
    has a trustline to the issuer for its `asset`, and the Stellar transaction completes
    successfully. All of these conditions and actions are mocked in this test to avoid
    network calls.
    """
    del mock_submit, mock_base_fee, client
    deposit = acc1_usd_deposit_transaction_factory()
    deposit.status = Transaction.STATUS.pending_anchor
    deposit.save()
    call_command("create_stellar_deposit", deposit.id)
    assert Transaction.objects.get(id=deposit.id).status == Transaction.STATUS.completed


@pytest.mark.django_db
@patch("stellar_sdk.server.Server.fetch_base_fee", return_value=100)
@patch("stellar_sdk.server.Server.submit_transaction", return_value=HORIZON_SUCCESS_RESPONSE)
@patch("polaris.helpers.check_auth", side_effect=mock_check_auth_success)
def test_deposit_interactive_confirm_success(
    mock_check,
    mock_submit,
    mock_base_fee,
    client,
    acc1_usd_deposit_transaction_factory,
):
    """
    `GET /deposit` and `GET /transactions/deposit/webapp` succeed with valid `account`
    and `asset_code`.
    """
    del mock_check, mock_submit, mock_base_fee
    deposit = acc1_usd_deposit_transaction_factory()
    response = client.post(
        DEPOSIT_PATH, {"asset_code": "USD", "account": deposit.stellar_account},
        follow=True
    )
    content = json.loads(response.content)
    assert response.status_code == 403
    assert content["type"] == "interactive_customer_info_needed"

    transaction_id = content["id"]
    url = content["url"]
    amount = 20
    response = client.post(url, {"amount": amount})
    assert response.status_code == 200
    assert (
        Transaction.objects.get(id=transaction_id).status
        == Transaction.STATUS.pending_user_transfer_start
    )

    response = client.get(
        f"/transactions/deposit/confirm_transaction?amount={amount}&transaction_id={transaction_id}",
        follow=True,
    )
    assert response.status_code == 200
    content = json.loads(response.content)
    transaction = content["transaction"]
    assert transaction

    # The transaction in the response was serialized before the Stellar transaction
    # to ensure deterministic behavior in testing. Thus, the response will contain
    # a transaction with status `pending_anchor`.
    assert transaction["status"] == Transaction.STATUS.pending_anchor
    assert float(transaction["amount_in"]) == amount

    # Since we have mocked the `create_deposit` function to success, the
    # transaction itself should be stored with status `completed`.
    assert (
        Transaction.objects.get(id=transaction_id).status
        == Transaction.STATUS.completed
    )


@pytest.mark.django_db
@patch("stellar_sdk.server.Server.fetch_base_fee", return_value=100)
@patch("stellar_sdk.server.Server.submit_transaction", return_value=HORIZON_SUCCESS_RESPONSE)
@patch(
    "stellar_sdk.call_builder.accounts_call_builder.AccountsCallBuilder.call",
    return_value={"sequence": 1, "balances": [{"asset_code": "USD"}]},
)
def test_deposit_check_trustlines_success(
    mock_account,
    mock_submit,
    mock_base_fee,
    client,
    acc1_usd_deposit_transaction_factory,
):
    """
    Creates a transaction with status `pending_trust` and checks that
    `check_trustlines` changes its status to `completed`. All the necessary
    functionality and conditions are mocked for determinism.
    """
    del mock_account, mock_submit, mock_base_fee, client
    deposit = acc1_usd_deposit_transaction_factory()
    deposit.status = Transaction.STATUS.pending_trust
    deposit.save()
    assert (
        Transaction.objects.get(id=deposit.id).status
        == Transaction.STATUS.pending_trust
    )
    call_command("check_trustlines")
    assert Transaction.objects.get(id=deposit.id).status == Transaction.STATUS.completed


@pytest.mark.django_db
@pytest.mark.skip
def test_deposit_check_trustlines_horizon(
    client, acc1_usd_deposit_transaction_factory
):
    """
    Tests the `check_trustlines` function's various logical paths. Note that the Stellar
    deposit is created synchronously. This makes Horizon calls, so it is skipped by the CI.
    """
    # Initiate a transaction with a new Stellar account.
    deposit = acc1_usd_deposit_transaction_factory()

    keypair = Keypair.random()
    deposit.stellar_account = keypair.public_key
    response = client.get(
        f"/deposit?asset_code=USD&account={deposit.stellar_account}", follow=True
    )
    content = json.loads(response.content)
    assert response.status_code == 403
    assert content["type"] == "interactive_customer_info_needed"

    # Complete the interactive deposit. The transaction should be set
    # to pending_user_transfer_start, since wallet-side confirmation has not happened.
    transaction_id = content["id"]
    url = content["url"]
    amount = 20
    response = client.post(url, {"amount": amount})
    assert response.status_code == 200
    assert (
        Transaction.objects.get(id=transaction_id).status
        == Transaction.STATUS.pending_user_transfer_start
    )

    # As a result of this external confirmation, the transaction should
    # be `pending_trust`. This will trigger a synchronous call to
    # `create_stellar_deposit`, which will register the account on testnet.
    # Since the account will not have a trustline, the status will still
    # be `pending_trust`.
    response = client.get(
        f"/transactions/deposit/confirm_transaction?amount={amount}&transaction_id={transaction_id}",
        follow=True,
    )
    assert response.status_code == 200
    content = json.loads(response.content)
    transaction = content["transaction"]
    assert transaction
    assert transaction["status"] == Transaction.STATUS.pending_anchor
    assert float(transaction["amount_in"]) == amount

    # The Stellar account has not been registered, so
    # this should not change the status of the Transaction.
    call_command("check_trustlines")
    assert (
        Transaction.objects.get(id=transaction_id).status
        == Transaction.STATUS.pending_trust
    )

    # Add a trustline for the transaction asset from the server
    # source account to the transaction account.
    from stellar_sdk.asset import Asset
    from stellar_sdk.transaction_builder import TransactionBuilder

    asset_code = deposit.asset.code
    asset_issuer = settings.STELLAR_DISTRIBUTION_ACCOUNT_ADDRESS
    Asset(code=asset_code, issuer=asset_issuer)

    server = settings.HORIZON_SERVER
    source_account = server.load_account(keypair.public_key)
    base_fee = 100
    transaction = TransactionBuilder(source_account=source_account,
                                     network_passphrase=settings.STELLAR_NETWORK_PASSPHRASE,
                                     base_fee=base_fee).append_change_trust_op(asset_code, asset_issuer).build()
    transaction.sign(keypair)
    response = server.submit_transaction(transaction)
    assert response["result_xdr"] == "AAAAAAAAAGQAAAAAAAAAAQAAAAAAAAAGAAAAAAAAAAA="

    call_command("check_trustlines")
    completed_transaction = Transaction.objects.get(id=transaction_id)
    assert completed_transaction.status == Transaction.STATUS.completed
    assert (
        completed_transaction.stellar_transaction_id == HORIZON_SUCCESS_RESPONSE["hash"]
    )


@pytest.mark.django_db
def test_deposit_authenticated_success(client, acc1_usd_deposit_transaction_factory):
    """`GET /deposit` succeeds with the SEP 10 authentication flow."""
    client_address = "GDKFNRUATPH4BSZGVFDRBIGZ5QAFILVFRIRYNSQ4UO7V2ZQAPRNL73RI"
    client_seed = "SDKWSBERDHP3SXW5A3LXSI7FWMMO5H7HG33KNYBKWH2HYOXJG2DXQHQY"
    deposit = acc1_usd_deposit_transaction_factory()

    # SEP 10.
    response = client.get(f"/auth?account={client_address}", follow=True)
    content = json.loads(response.content)

    envelope_xdr = content["transaction"]
    envelope_object = TransactionEnvelope.from_xdr(envelope_xdr, network_passphrase=settings.STELLAR_NETWORK_PASSPHRASE)
    client_signing_key = Keypair.from_secret(client_seed)
    envelope_object.sign(client_signing_key)
    client_signed_envelope_xdr = envelope_object.to_xdr()

    response = client.post(
        "/auth",
        data={"transaction": client_signed_envelope_xdr},
        content_type="application/json",
    )
    content = json.loads(response.content)
    encoded_jwt = content["token"]
    assert encoded_jwt

    header = {"HTTP_AUTHORIZATION": f"Bearer {encoded_jwt}"}
    response = client.post(
        DEPOSIT_PATH, {"asset_code": "USD", "account": deposit.stellar_account},
        follow=True,
        **header,
    )
    content = json.loads(response.content)
    assert response.status_code == 403
    assert content["type"] == "interactive_customer_info_needed"


@pytest.mark.django_db
def test_deposit_no_jwt(client, acc1_usd_deposit_transaction_factory):
    """`GET /deposit` fails if a required JWT isn't provided."""
    deposit = acc1_usd_deposit_transaction_factory()
    response = client.post(
        DEPOSIT_PATH, {"asset_code": "USD", "account": deposit.stellar_account, "memo_type": "text", "memo": "foo"},
        follow=True,
    )
    content = json.loads(response.content)
    assert response.status_code == 400
    assert content == {"error": "JWT must be passed as 'Authorization' header", "status_code": 400}



from django.contrib import admin
from polaris.models import Transaction, Asset


class TransactionAdmin(admin.ModelAdmin):
    """
    This defines the admin view of a Transaction.
    """

    list_display = "id", "asset_name", "kind", "status", "started_at"


class AssetAdmin(admin.ModelAdmin):
    """
    This defines the admin view of an Asset.
    """

    list_display = "code", "issuer", "deposit_enabled", "withdrawal_enabled"


admin.site.register(Transaction, TransactionAdmin)
admin.site.register(Asset, AssetAdmin)



"""This module defines the URL patterns for the `/deposit` endpoint."""
from django.urls import path
from django.views.decorators.csrf import csrf_exempt
from polaris.deposit.views import deposit, interactive_deposit, confirm_transaction

urlpatterns = [
    path("transactions/deposit/interactive", csrf_exempt(deposit)),
    path("transactions/deposit/webapp", interactive_deposit, name="interactive_deposit"),
    path("transactions/deposit/confirm_transaction", confirm_transaction, name="confirm_transaction"),
]



"""This module defines the `/fee` view."""
from polaris import settings
from rest_framework.decorators import api_view
from rest_framework.response import Response

from polaris.helpers import calc_fee, render_error_response, validate_sep10_token
from polaris.models import Asset

OPERATION_DEPOSIT = settings.OPERATION_DEPOSIT
OPERATION_WITHDRAWAL = settings.OPERATION_WITHDRAWAL


def _op_type_is_valid(asset_code: str, operation: str, op_type: str) -> bool:
    asset = Asset.objects.get(code=asset_code)
    if all([operation == OPERATION_WITHDRAWAL, asset.withdrawal_enabled, not op_type]):
        return True

    if all([operation == OPERATION_DEPOSIT, asset.deposit_enabled, not op_type]):
        return True

    return False


@api_view()
@validate_sep10_token()
def fee(request):
    """
    Definition of the /fee endpoint, in accordance with SEP-0024.
    See: https://github.com/stellar/stellar-protocol/blob/master/ecosystem/sep-0024.md#fee
    """
    # Verify that the asset code exists in our database:
    asset_code = request.GET.get("asset_code")
    if not asset_code or not Asset.objects.filter(code=asset_code).exists():
        return render_error_response("invalid 'asset_code'")
    asset = Asset.objects.get(code=asset_code)

    # Verify that the requested operation is valid:
    operation = request.GET.get("operation")
    if operation not in (OPERATION_DEPOSIT, OPERATION_WITHDRAWAL):
        return render_error_response(
            f"'operation' should be either '{OPERATION_DEPOSIT}' or '{OPERATION_WITHDRAWAL}'"
        )
    # Verify that amount is provided, and that it is parseable into a float:
    amount_str = request.GET.get("amount")
    try:
        amount = float(amount_str)
    except (TypeError, ValueError):
        return render_error_response("invalid 'amount'")

    # Validate that the operation, and the specified type (if provided)
    # are applicable to the given asset:
    op_type = request.GET.get("type", "")
    if not _op_type_is_valid(asset_code, operation, op_type):
        return render_error_response(
            f"the specified operation is not available for '{asset_code}'"
        )

    return Response({"fee": calc_fee(asset, operation, amount)})



"""This module defines the URL patterns for the `/withdraw` endpoint."""
from django.urls import path
from django.views.decorators.csrf import csrf_exempt
from polaris.withdraw.views import withdraw, interactive_withdraw

urlpatterns = [
    path("transactions/withdraw/interactive", csrf_exempt(withdraw)),
    path("transactions/withdraw/webapp", interactive_withdraw, name="interactive_withdraw"),
]



"""This module tests the <auth> endpoint."""
import json

from stellar_sdk.keypair import Keypair
from stellar_sdk.transaction_envelope import TransactionEnvelope
from stellar_sdk.xdr import Xdr

from polaris import settings
from polaris.tests.conftest import STELLAR_ACCOUNT_1

CLIENT_ADDRESS = "GDKFNRUATPH4BSZGVFDRBIGZ5QAFILVFRIRYNSQ4UO7V2ZQAPRNL73RI"
CLIENT_SEED = "SDKWSBERDHP3SXW5A3LXSI7FWMMO5H7HG33KNYBKWH2HYOXJG2DXQHQY"


def test_auth_get_no_account(client):
    """`GET <auth>` fails with no `account` parameter."""
    response = client.get("/auth", follow=True)
    content = json.loads(response.content)
    assert response.status_code == 400
    assert content == {"error": "no 'account' provided"}


def test_auth_get_account(client):
    """`GET <auth>` succeeds with a valid TransactionEnvelope XDR."""
    response = client.get(f"/auth?account={STELLAR_ACCOUNT_1}", follow=True)
    content = json.loads(response.content)
    assert content["network_passphrase"] == "Test SDF Network ; September 2015"
    assert content["transaction"]

    envelope_xdr = content["transaction"]
    envelope_object = TransactionEnvelope.from_xdr(envelope_xdr, network_passphrase=settings.STELLAR_NETWORK_PASSPHRASE)
    transaction_object = envelope_object.transaction
    assert transaction_object.sequence == 0
    assert len(transaction_object.operations) == 1

    manage_data_op = transaction_object.operations[0]
    assert manage_data_op.type_code() == Xdr.const.MANAGE_DATA
    assert manage_data_op.data_name == "SEP 24 Reference auth"
    assert len(manage_data_op.data_value) == 64

    signatures = envelope_object.signatures
    assert len(signatures) == 1
    server_signature = signatures[0]

    tx_hash = envelope_object.hash()
    server_public_key = Keypair.from_public_key(settings.STELLAR_DISTRIBUTION_ACCOUNT_ADDRESS)
    server_public_key.verify(tx_hash, server_signature.signature)


def test_auth_post_json_success(client):
    """`POST <auth>` succeeds when given a proper JSON-encoded transaction."""
    response = client.get(f"/auth?account={CLIENT_ADDRESS}", follow=True)
    content = json.loads(response.content)

    # Sign the XDR with the client.
    envelope_xdr = content["transaction"]
    envelope_object = TransactionEnvelope.from_xdr(envelope_xdr, network_passphrase=settings.STELLAR_NETWORK_PASSPHRASE)
    client_signing_key = Keypair.from_secret(CLIENT_SEED)
    envelope_object.sign(client_signing_key)
    client_signed_envelope_xdr = envelope_object.to_xdr()

    response = client.post(
        "/auth",
        data={"transaction": client_signed_envelope_xdr},
        content_type="application/json",
    )

    content = json.loads(response.content)
    assert content["token"]


def test_auth_post_urlencode_success(client):
    """`POST <auth>` succeeds when given a proper URL-encoded transaction."""
    response = client.get(f"/auth?account={CLIENT_ADDRESS}", follow=True)
    content = json.loads(response.content)

    # Sign the XDR with the client.
    envelope_xdr = content["transaction"]
    envelope_object = TransactionEnvelope.from_xdr(envelope_xdr, network_passphrase=settings.STELLAR_NETWORK_PASSPHRASE)
    client_signing_key = Keypair.from_secret(CLIENT_SEED)
    envelope_object.sign(client_signing_key)
    client_signed_envelope_xdr = envelope_object.to_xdr()

    response = client.post(
        "/auth",
        data=f"transaction=<{client_signed_envelope_xdr}>",
        content_type="application/x-www-form-urlencoded",
    )
    content = json.loads(response.content)
    assert content["token"]



"""This module tests the `/info` endpoint."""
import json
import pytest


def _get_expected_response(settings):
    """
    This expected response was adapted from the example SEP-0024 response on
    https://github.com/stellar/stellar-protocol/blob/master/ecosystem/sep-0024.md#response-2
    Some changes have been applied, to ensure the data we provide is in a consistent format and
    in accordance with design decisions from this reference implementation:

    - The "optional" configuration is always explicit, to avoid misinterpretation
    - All SEP24 endpoints are enabled=true
    - SEP 24 requires all deposits and withdrawals to have SEP 10 authentication
    - All floating numbers are displayed with at least one decimal
    - min_amount and max_amount are mandatorily informed
    """

    return f"""
    {{
        "deposit": {{
            "USD": {{
                "enabled": true,
                "fee_fixed": 5.0,
                "fee_percent": 1.0,
                "min_amount": 0.1,
                "max_amount": 1000.0
            }},
            "ETH": {{
                "enabled": true,
                "fee_fixed": 0.002,
                "fee_percent": 0.0,
                "max_amount": 10000000.0,
                "min_amount": 0.0
            }}
        }},
        "withdraw": {{
            "USD": {{
                "enabled": true,
                "fee_fixed": 5.0,
                "fee_percent": 0.0,
                "min_amount": 0.1,
                "max_amount": 1000.0
            }},
            "ETH": {{"enabled": false}}
        }},
        "fee": {{"enabled": true}},
        "transactions": {{"enabled": true}},
        "transaction": {{"enabled": true}}
    }}"""


@pytest.mark.django_db
def test_info_endpoint(client, settings, usd_asset_factory, eth_asset_factory):
    """
    Ensures the /info endpoint provides data in the expected format, according to
    SEP 24: https://github.com/stellar/stellar-protocol/blob/master/ecosystem/sep-0024.md
    """
    usd_asset_factory()
    eth_asset_factory()

    response = client.get(f"/info", follow=True)
    content = json.loads(response.content)
    expected_response = json.loads(_get_expected_response(settings))
    assert content == expected_response
    assert response.status_code == 200



"""This module defines forms used in the withdraw view."""
from django import forms


class WithdrawForm(forms.Form):
    """This form accepts the amount to withdraw from the user."""

    amount = forms.FloatField(
        help_text="Enter the amount to withdraw, as a two decimal places float.",
        min_value=0,
        widget=forms.NumberInput(attrs={"class": "input"}),
    )
    asset = None
    bank_account = forms.CharField(
        help_text="Enter the bank account number for withdrawal.", min_length=0
    )
    # TODO: Replace the bank with a ChoiceField.
    bank = forms.CharField(help_text="Enter the bank to withdraw from.", min_length=0)

    def clean_amount(self):
        """Validate the provided amount of an asset."""
        amount = round(self.cleaned_data["amount"], 2)
        if self.asset:
            if amount < self.asset.withdrawal_min_amount:
                raise forms.ValidationError(
                    f"Amount is below minimum for asset {self.asset.code}"
                )
            elif amount > self.asset.withdrawal_max_amount:
                raise forms.ValidationError(
                    f"Amount is above maximum for asset {self.asset.code}"
                )
        return amount



"""
Django settings for app project.

Generated by 'django-admin startproject' using Django 2.2.2.

For more information on this file, see
https://docs.djangoproject.com/en/2.2/topics/settings/

For the full list of settings and their values, see
https://docs.djangoproject.com/en/2.2/ref/settings/
"""
# pylint: disable=invalid-name
import os
from shutil import copyfile

import environ
from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from stellar_sdk.server import Server
from stellar_sdk.keypair import Keypair


# Build paths inside the project like this: os.path.join(BASE_DIR, ...)

BASE_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))

# Load environment variables from .env
env = environ.Env()

# Use the directory of the django project that installed this app
# Or if undefined, use outer directory for this app
try:
    PROJECT_ROOT = settings.PROJECT_ROOT
except ImproperlyConfigured:
    PROJECT_ROOT = os.path.dirname(BASE_DIR)
    copyfile(os.path.join(PROJECT_ROOT, ".env.example"), os.path.join(PROJECT_ROOT, ".env"))

env_file = os.path.join(PROJECT_ROOT, ".env")
if os.path.exists(env_file):
    environ.Env.read_env(str(env_file))

SECRET_KEY = env("DJANGO_SECRET_KEY")
DEBUG = env.bool("DJANGO_DEBUG", False)

STELLAR_DISTRIBUTION_ACCOUNT_SEED = env("STELLAR_DISTRIBUTION_ACCOUNT_SEED")
STELLAR_DISTRIBUTION_ACCOUNT_ADDRESS = (
    Keypair.from_secret(STELLAR_DISTRIBUTION_ACCOUNT_SEED).public_key
)
STELLAR_ISSUER_ACCOUNT_ADDRESS = env("STELLAR_ISSUER_ACCOUNT_ADDRESS")
STELLAR_NETWORK_PASSPHRASE = env("STELLAR_NETWORK_PASSPHRASE", default="Test SDF Network ; September 2015")
HORIZON_URI = env("HORIZON_URI", default="https://horizon-testnet.stellar.org/")
HORIZON_SERVER = Server(horizon_url=HORIZON_URI)
SERVER_JWT_KEY = env("SERVER_JWT_KEY")
OPERATION_DEPOSIT = "deposit"
OPERATION_WITHDRAWAL = "withdraw"
ACCOUNT_STARTING_BALANCE = str(2.01)

# Apps to add to parent project's INSTALLED_APPS
django_apps = [
    "django.contrib.admin",
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sessions",
    "django.contrib.messages",
    "django.contrib.staticfiles",
]
third_party_apps = ["rest_framework", "corsheaders"]
INSTALLED_APPS = django_apps + third_party_apps + ["polaris"]

# Modules to add to parent project's MIDDLEWARE
MIDDLEWARE = [
    "corsheaders.middleware.CorsMiddleware",
    "whitenoise.middleware.WhiteNoiseMiddleware",
]

APPEND_SLASH = False

ROOT_URLCONF = "polaris.urls"

TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "DIRS": [os.path.join(BASE_DIR, "templates")],
        "APP_DIRS": True,
        "OPTIONS": {
            "context_processors": [
                "django.template.context_processors.debug",
                "django.template.context_processors.request",
                "django.contrib.auth.context_processors.auth",
                "django.contrib.messages.context_processors.messages",
            ]
        },
    }
]

# Database
# https://docs.djangoproject.com/en/2.2/ref/settings/#databases

DATABASES = {
    "default": env.db(
        "DATABASE_URL", default="sqlite:///" + os.path.join(PROJECT_ROOT, "db.sqlite3")
    )
}

# Static files (CSS, JavaScript, Images)
# https://docs.djangoproject.com/en/2.2/howto/static-files/

STATIC_ROOT = os.path.join(BASE_DIR, "polaris/static")
STATIC_URL = "/polaris/static/"
STATICFILES_STORAGE = "whitenoise.storage.CompressedManifestStaticFilesStorage"
STATICFILES_DIRS = (
    os.path.join(BASE_DIR, 'polaris/static'),
)

# Django Rest Framework Settings:
# Attributes to add to parent project's REST_FRAMEWORK
REST_FRAMEWORK = {
    "DEFAULT_PAGINATION_CLASS": "rest_framework.pagination.PageNumberPagination",
    "DEFAULT_RENDERER_CLASSES": [
        'rest_framework.renderers.JSONRenderer',
        'rest_framework.renderers.BrowsableAPIRenderer',
        'rest_framework.renderers.TemplateHTMLRenderer'
    ],
    "PAGE_SIZE": 10,
}

# API Config

DEFAULT_PAGE_SIZE = 10

# CORS configuration

CORS_ORIGIN_ALLOW_ALL = True



import time
import logging

from polaris import settings
from django.core.management.base import BaseCommand, CommandError
from django.core.management import call_command
from stellar_sdk.exceptions import BaseHorizonError

from polaris.models import Transaction

logger = logging.getLogger(__name__)


class Command(BaseCommand):
    """
    Create Stellar transaction for deposit transactions marked as pending trust, if a
    trustline has been created.
    """
    def add_arguments(self, parser):
        parser.add_argument("--loop", "-l", action="store_true")

    def handle(self, *args, **options):
        if options.get('loop'):
            while True:
                self.check_trustlines()
                time.sleep(60)
        else:
            self.check_trustlines()

    def check_trustlines(self):
        """
        Create Stellar transaction for deposit transactions marked as pending trust, if a
        trustline has been created.
        """
        transactions = Transaction.objects.filter(status=Transaction.STATUS.pending_trust)
        server = settings.HORIZON_SERVER
        for transaction in transactions:
            try:
                account = server.accounts().account_id(transaction.stellar_account).call()
            except BaseHorizonError:
                logger.debug("could not load account using provided horizon URL")
                continue
            try:
                balances = account["balances"]
            except KeyError:
                logger.debug("horizon account response had no balances")
                continue
            for balance in balances:
                try:
                    asset_code = balance["asset_code"]
                except KeyError:
                    logger.debug("horizon balances had no asset_code")
                    continue
                if asset_code == transaction.asset.code:
                    call_command('create_stellar_deposit', transaction.id)




"""This module defines custom management commands for the app admin."""
import logging

from django.core.management.base import BaseCommand, CommandError
from django.utils.timezone import now
from stellar_sdk.exceptions import NotFoundError
from stellar_sdk.transaction_envelope import TransactionEnvelope
from stellar_sdk.xdr import Xdr

from polaris import settings
from polaris.helpers import format_memo_horizon
from polaris.models import Transaction

logger = logging.getLogger(__file__)


def stream_transactions():
    """Stream transactions for the server Stellar address. Decomposed for easier testing."""
    server = settings.HORIZON_SERVER
    try:
        # Ensure the distribution account actually exists
        server.load_account(settings.STELLAR_DISTRIBUTION_ACCOUNT_ADDRESS)
    except NotFoundError as exc:
        raise RuntimeError("Stellar distribution account does not exist in horizon")
    return server.transactions().for_account(settings.STELLAR_DISTRIBUTION_ACCOUNT_ADDRESS).cursor("now").stream()


def _check_payment_op(operation, want_asset, want_amount):
    if operation.type_code() != Xdr.const.PAYMENT:
        return False
    if str(operation.destination) != settings.STELLAR_DISTRIBUTION_ACCOUNT_ADDRESS:
        return False
    if str(operation.asset.code) != want_asset:
        return False
    # TODO: Handle multiple possible asset issuance accounts
    if str(operation.asset.issuer) != settings.STELLAR_ISSUER_ACCOUNT_ADDRESS:
        return False
    if float(operation.amount) != want_amount:
        return False
    return True


def process_withdrawal(response, transaction):
    """
    Check if a Stellar transaction's memo matches the ID of a database transaction.
    """
    # Validate the Horizon response.
    try:
        memo_type = response["memo_type"]
        response_memo = response["memo"]
        successful = response["successful"]
        stellar_transaction_id = response["id"]
        envelope_xdr = response["envelope_xdr"]
    except KeyError:
        return False

    if memo_type != "hash":
        return False

    # The memo on the response will be base 64 string, due to XDR, while
    # the memo parameter is base 16. Thus, we convert the parameter
    # from hex to base 64, and then to a string without trailing whitespace.
    if response_memo != format_memo_horizon(transaction.withdraw_memo):
        return False

    horizon_tx = TransactionEnvelope.from_xdr(envelope_xdr, network_passphrase=settings.STELLAR_NETWORK_PASSPHRASE).transaction
    found_matching_payment_op = False
    for operation in horizon_tx.operations:
        if _check_payment_op(operation, transaction.asset.code, transaction.amount_in):
            found_matching_payment_op = True
            break

    # If no matching payment operation is found in the Stellar response, return.
    if not found_matching_payment_op:
        return False

    # If the Stellar transaction succeeded, we mark the corresponding `Transaction`
    # accordingly in the database. Else, we mark it `pending_stellar`, so the wallet
    # knows to resubmit.
    if successful:
        transaction.completed_at = now()
        transaction.status = Transaction.STATUS.completed
        transaction.status_eta = 0
        transaction.amount_out = transaction.amount_in - transaction.amount_fee
    else:
        transaction.status = Transaction.STATUS.pending_stellar

    transaction.stellar_transaction_id = stellar_transaction_id
    transaction.save()
    return True


class Command(BaseCommand):
    """
    Custom command to monitor Stellar transactions to the anchor account.
    It also updates the transactions in the database.
    """

    help = "Watches for Stellar transactions to the anchor account using Horizon"

    def handle(self, *args, **options):
        transaction_responses = stream_transactions()
        for response in transaction_responses:
            pending_withdrawal_transactions = Transaction.objects.filter(
                status=Transaction.STATUS.pending_user_transfer_start
            ).filter(kind=Transaction.KIND.withdrawal)
            for withdrawal_transaction in pending_withdrawal_transactions:
                if process_withdrawal(response, withdrawal_transaction):
                    envelope_xdr = response["envelope_xdr"]
                    logger.info(
                        f"successfully processed withdrawal for response with xdr {envelope_xdr}"
                    )
                    break



"""This module registers the models for the `info` endpoint with the admin."""
from django.contrib import admin
from polaris.models import Asset


class AssetAdmin(admin.ModelAdmin):
    """
    This defines the admin view of an Asset.
    """

    list_display = "code", "issuer", "deposit_enabled", "withdrawal_enabled"


admin.site.register(Asset, AssetAdmin)



from setuptools import setup, find_packages


with open("README.md") as f:
    long_description = f.read()

setup(
    name="django-polaris",
    version="0.9.2",
    description="A SEP-24-compliant Django anchor server",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/stellar/django-polaris",
    author="Jake Urban",
    author_email="jake@stellar.org",
    license="Apache license 2.0",
    classifiers=[
        "Environment :: Web Environment",
        "Framework :: Django",
        "Framework :: Django :: 2.2",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.7",
    ],
    keywords=["stellar", "sdf", "anchor", "server", "polaris", "sep-24", "sep24"],
    include_package_data=True,
    package_dir={"": "polaris"},
    packages=find_packages("polaris"),
    install_requires=[
        "aiohttp==4.0.0a1",
        "aiohttp-sse-client==0.1.6",
        "async-timeout==3.0.1",
        "attrs==19.3.0",
        "certifi==2019.9.11",
        "cffi==1.13.2",
        "chardet==3.0.4",
        "crc16==0.1.1",
        "django==2.2.4",
        "django-cors-headers==3.2.0",
        "django-environ==0.4.5",
        "django-model-utils==3.2.0",
        "djangorestframework==3.10.3",
        "gunicorn==20.0.0",
        "idna==2.8",
        "mnemonic==0.19",
        "multidict==4.6.0a2",
        "psycopg2-binary==2.8.4",
        "pycparser==2.19",
        "pyjwt==1.7.1",
        "pynacl==1.3.0",
        "pytz==2019.3",
        "requests==2.22.0",
        "six==1.13.0",
        "sqlparse==0.3.0",
        "stellar-base-sseclient==0.0.21",
        "stellar-sdk==2.0.0b5",
        "toml==0.10.0",
        "typing-extensions==3.7.4.1",
        "urllib3==1.25.7",
        "whitenoise==4.1.4",
        "yarl==1.4.0a11",
    ],
    python_requires=">=3.7",
)



"""This module tests the `/fee` endpoint. All the below tests call `GET /fee`."""
import json
from unittest.mock import patch

import pytest
from stellar_sdk.keypair import Keypair
from stellar_sdk.transaction_envelope import TransactionEnvelope

from polaris import settings
from polaris.tests.helpers import mock_check_auth_success


@patch("polaris.helpers.check_auth", side_effect=mock_check_auth_success)
def test_fee_no_params(mock_check, client):
    """Fails with no params provided."""
    del mock_check
    response = client.get(f"/fee", follow=True)
    content = json.loads(response.content)

    assert response.status_code == 400
    assert content == {"error": "invalid 'asset_code'", "status_code": 400}


@pytest.mark.django_db
@patch("polaris.helpers.check_auth", side_effect=mock_check_auth_success)
def test_fee_wrong_asset_code(mock_check, client):
    """Fails with an invalid `asset_code`."""
    del mock_check
    response = client.get(f"/fee?asset_code=NADA", follow=True)
    content = json.loads(response.content)

    assert response.status_code == 400
    assert content == {"error": "invalid 'asset_code'", "status_code": 400}


@pytest.mark.django_db
@patch("polaris.helpers.check_auth", side_effect=mock_check_auth_success)
def test_fee_no_operation(mock_check, client, usd_asset_factory):
    """Fails with no `operation` provided."""
    del mock_check
    usd_asset_factory()
    response = client.get(f"/fee?asset_code=USD", follow=True)
    content = json.loads(response.content)

    assert response.status_code == 400
    assert content == {
        "error": "'operation' should be either 'deposit' or 'withdraw'",
        "status_code": 400
    }


@pytest.mark.django_db
@patch("polaris.helpers.check_auth", side_effect=mock_check_auth_success)
def test_fee_invalid_operation(mock_check, client, usd_asset_factory):
    """Fails with an invalid `operation` provided."""
    del mock_check
    usd_asset_factory()
    response = client.get(f"/fee?asset_code=USD&operation=generate", follow=True)
    content = json.loads(response.content)

    assert response.status_code == 400
    assert content == {
        "error": "'operation' should be either 'deposit' or 'withdraw'",
        "status_code": 400
    }


@pytest.mark.django_db
@patch("polaris.helpers.check_auth", side_effect=mock_check_auth_success)
def test_fee_no_amount(mock_check, client, usd_asset_factory):
    """Fails with no amount provided."""
    del mock_check
    usd_asset_factory()
    response = client.get(f"/fee?asset_code=USD&operation=withdraw", follow=True)
    content = json.loads(response.content)

    assert response.status_code == 400
    assert content == {
        "error": "invalid 'amount'",
        "status_code": 400
    }


@pytest.mark.django_db
@patch("polaris.helpers.check_auth", side_effect=mock_check_auth_success)
def test_fee_invalid_amount(mock_check, client, usd_asset_factory):
    """Fails with a non-float amount provided."""
    del mock_check
    usd_asset_factory()
    response = client.get(
        f"/fee?asset_code=USD&operation=withdraw&amount=TEXT", follow=True
    )
    content = json.loads(response.content)

    assert response.status_code == 400
    assert content == {"error": "invalid 'amount'", "status_code": 400}


@pytest.mark.django_db
@patch("polaris.helpers.check_auth", side_effect=mock_check_auth_success)
def test_fee_invalid_operation_type_deposit(mock_check, client, usd_asset_factory):
    """Fails if the specified deposit `operation` is not valid for `asset_code`."""
    del mock_check
    usd_asset_factory()
    response = client.get(
        f"/fee?asset_code=USD&operation=deposit&amount=100.0&type=IBAN", follow=True
    )
    content = json.loads(response.content)

    assert response.status_code == 400
    assert content == {
        "error": "the specified operation is not available for 'USD'",
        "status_code": 400
    }


@pytest.mark.django_db
@patch("polaris.helpers.check_auth", side_effect=mock_check_auth_success)
def test_fee_invalid_operation_type_withdraw(mock_check, client, usd_asset_factory):
    """Fails if the specified withdraw `operation` is not enabled for `asset_code`."""
    del mock_check
    usd_asset_factory()
    response = client.get(
        f"/fee?asset_code=USD&operation=withdraw&amount=100.0&type=IBAN", follow=True
    )
    content = json.loads(response.content)

    assert response.status_code == 400
    assert content == {
        "error": "the specified operation is not available for 'USD'",
        "status_code": 400
    }


@pytest.mark.django_db
@patch("polaris.helpers.check_auth", side_effect=mock_check_auth_success)
def test_fee_withdraw_disabled(mock_check, client, eth_asset_factory):
    """Fails if the withdraw `operation` is not enabled for the `asset_code`."""
    del mock_check
    eth_asset_factory()

    response = client.get(
        f"/fee?asset_code=ETH&operation=withdraw&amount=100.0", follow=True
    )
    content = json.loads(response.content)

    assert response.status_code == 400
    assert content == {
        "error": "the specified operation is not available for 'ETH'",
        "status_code": 400
    }


@pytest.mark.django_db
@patch("polaris.helpers.check_auth", side_effect=mock_check_auth_success)
def test_fee_deposit_disabled(mock_check, client, eth_asset_factory):
    """Fails if the deposit `operation` is not enabled for `asset_code`."""
    del mock_check
    asset = eth_asset_factory()
    asset.deposit_enabled = False
    asset.save()

    response = client.get(
        f"/fee?asset_code=ETH&operation=deposit&amount=100.0", follow=True
    )
    content = json.loads(response.content)

    assert response.status_code == 400
    assert content == {
        "error": "the specified operation is not available for 'ETH'",
        "status_code": 400
    }


@pytest.mark.django_db
@patch("polaris.helpers.check_auth", side_effect=mock_check_auth_success)
def test_fee_valid_deposit(mock_check, client, usd_asset_factory):
    """Succeeds for a valid deposit."""
    del mock_check
    usd_asset_factory()

    response = client.get(
        f"/fee?asset_code=USD&operation=deposit&amount=200.0", follow=True
    )
    content = json.loads(response.content)

    assert response.status_code == 200
    assert content == {"fee": 5.0 + 2.0}


# Fixed: 5.0 Percent = 1
@pytest.mark.django_db
@patch("polaris.helpers.check_auth", side_effect=mock_check_auth_success)
def test_fee_valid_withdrawal(mock_check, client, usd_asset_factory):
    """Succeeds for a valid withdrawal."""
    del mock_check
    usd_asset_factory()

    response = client.get(
        f"/fee?asset_code=USD&operation=withdraw&amount=100.0", follow=True
    )
    content = json.loads(response.content)

    assert response.status_code == 200
    assert content == {"fee": 5.0}


@pytest.mark.django_db
def test_fee_authenticated_success(client, usd_asset_factory):
    """Succeeds for a valid fee, with successful authentication."""
    usd_asset_factory()
    client_address = "GDKFNRUATPH4BSZGVFDRBIGZ5QAFILVFRIRYNSQ4UO7V2ZQAPRNL73RI"
    client_seed = "SDKWSBERDHP3SXW5A3LXSI7FWMMO5H7HG33KNYBKWH2HYOXJG2DXQHQY"

    # SEP 10.
    response = client.get(f"/auth?account={client_address}", follow=True)
    content = json.loads(response.content)
    envelope_xdr = content["transaction"]
    envelope_object = TransactionEnvelope.from_xdr(envelope_xdr, network_passphrase=settings.STELLAR_NETWORK_PASSPHRASE)
    client_signing_key = Keypair.from_secret(client_seed)
    envelope_object.sign(client_signing_key)
    client_signed_envelope_xdr = envelope_object.to_xdr()

    response = client.post(
        "/auth",
        data={"transaction": client_signed_envelope_xdr},
        content_type="application/json",
    )
    content = json.loads(response.content)
    encoded_jwt = content["token"]
    assert encoded_jwt

    # For testing, we make the key `HTTP_AUTHORIZATION`. This is the value that
    # we expect due to the middleware.
    header = {"HTTP_AUTHORIZATION": f"Bearer {encoded_jwt}"}

    response = client.get(
        f"/fee?asset_code=USD&operation=withdraw&amount=100.0", follow=True, **header
    )
    content = json.loads(response.content)

    assert response.status_code == 200
    assert content == {"fee": 5.0}


@pytest.mark.django_db
def test_fee_no_jwt(client, usd_asset_factory):
    """`GET /fee` fails if a required JWT is not provided."""
    usd_asset_factory()
    response = client.get(
        f"/fee?asset_code=USD&operation=withdraw&amount=100.0", follow=True
    )
    content = json.loads(response.content)
    assert response.status_code == 400
    assert content == {
        "error": "JWT must be passed as 'Authorization' header",
        "status_code": 400
    }
